from src.FluxUpMaterials import FluxUpMaterials
from src.FluxUpDepletionResults import FluxUpDepletionResults
import openmc
import numpy as np
import src.tallies
from abc import ABC as AbstractBaseClass, abstractmethod
import pandas as pd
import src.utilities as ut


class CardContainer(object):
    '''
    Container class for "cards" - geometry, universe and cell "cards" in MCNP/Serpent parlance.
    '''
    pass

class AssemblyModel(AbstractBaseClass):
    '''
    Abstract base class for assembly models.
    
    Changes
    ----------
    01 Dec 2023 Andy - Add more type hinting.
    '''

    def __init__(self):
        self.materials = FluxUpMaterials()
        self.populateMaterials()

    def populateDepletedMaterials(self,depletedMaterialsDirectories:list,burnableMatID:int) -> None:
        self.fuelMaterialIDs = []
        for dir in depletedMaterialsDirectories:
            depletionResults = FluxUpDepletionResults(dir,crossSectionsXmlFullPath=openmc.Materials.cross_sections)
            # the fuelMaterialIDs in this case is a 2-D list - in the self.materials object however, the IDs are stored in a flat list.
            # this is deliberate, to make it easier/more flexible to organize fuel IDs, since we tend to tabulate fuel in terms of
            # enrichment and burnup.
            self.fuelMaterialIDs.append(self.materials.appendFromDepletionResults(depletionResults,burnableMatID))

    '''
    Abstract methods: subclasses must implement
    '''

    @abstractmethod
    def populateMaterials(self):
        pass

    @abstractmethod
    def computeGeometry(self):
        pass
    
    @abstractmethod
    def createFullUniverse(self):
        pass

    @abstractmethod
    def settings(self):
        pass

    @abstractmethod
    def tallies(self):
        pass        

    def openmcModel(self, sampleData:pd.DataFrame, customData:list=None) -> openmc.model.Model:
        '''
        Return a fully initialized OpenMC assembly model. This function should be set in OpenMCWrapper.customModelFunctions.

        Parameters
        ----------
            sampleData: DataFrame
                Pandas DataFrame row containing assembly parameters.
            customData: Dictionary
                Dictionary containing any additional data needed to instantiate the OpenMC model. In this class's implementation
                of openmcModel() this is unused, but the input parameter is kept for compatibility.
                
        Returns
        ----------
            model: openmc.model.Model
                An OpenMC Model

        Notes:
        ----------
        '''
        geom = self.computeGeometry(sampleData)       
        fullUniverse = self.createFullUniverse(sampleData, geom)
        openmcGeom = openmc.Geometry(fullUniverse)
        model = openmc.model.Model(geometry=openmcGeom,materials=self.materials,settings=self.settings(geom),tallies=self.tallies(geom),plots=self.plots())
        return model,geom

    def plots(self):
        return None

    


class AssemblyModelSFR(AssemblyModel):
    '''
    Changes
    ----------
    01 Dec 2023 Andy - Add more type hinting.
    '''

    def __init__(self):
        super().__init__()
        

    def populateMaterials(self):
        self.claddingMaterialID = self.materials.append(FluxUpMaterials.claddingHT9())
        self.coolantMaterialID  = self.materials.append(FluxUpMaterials.coolantNa())
        #self.coolantMaterialID  = self.materials.append(FluxUpMaterials.coolantH2O())
        self.absorberMaterialID = self.materials.append(FluxUpMaterials.absorberEBRII())
        self.fuelMaterialID     = self.materials.append(FluxUpMaterials.fuelESFR())
        #self.fuelMaterialID     = self.materials.append(FluxUpMaterials.fuelUO2_9ao())

    def getCoolantMaterial(self):
        return self.materials.getByID(self.coolantMaterialID)

    def getCladdingMaterial(self):
        return self.materials.getByID(self.claddingMaterialID)

    def getAbsorberMaterial(self):
        return self.materials.getByID(self.absorberMaterialID)

    def getMaterialByID(self,matID:int) -> openmc.Material:
        return self.materials.getByID(matID)

    def computeGeometry(self,sampleData:pd.DataFrame) -> CardContainer:
        '''
        Parameters
        ----------
            sampleData: DataFrame
                Pandas DataFrame row containing assembly parameters.
           
        Returns
        ----------
            cont: CardContainer
                Instance of a CardContainer holding onto geometric quantities

        Changes
        ----------
        01 Dec 2023 Andy - define additional variables for clarity        
        '''
        geom = CardContainer()
       # hex assembly lattice ("large" lattice of hex assemblies)
        longDiagonal = (2*sampleData['num_rows'] -1)*sampleData['pin_pitch'] # fuel-to-fuel
        assemblyInnerPadding = sampleData['pin_pitch'] 
        geom.assemblyFlatToFlatInner = np.sqrt(3)*longDiagonal/2 + assemblyInnerPadding # flat-to-flat length: inside of hex cylinder only
        geom.assemblyPitch =  geom.assemblyFlatToFlatInner + 2*sampleData['duct_thickness'] # overall flat-to-flat assembly length
        geom.cladOuterRadius = sampleData['pin_radius']
        geom.fuelOuterRadius = geom.cladOuterRadius - sampleData['clad_thickness']
        geom.cladInnerRadius = geom.fuelOuterRadius
        
        geom.assemblySideLength = geom.assemblyFlatToFlatInner/np.sqrt(3)
        #geom.boundingBoxSideLength = geom.assemblyFlatToFlatInner*2
        geom.boundingBoxWidth = np.sqrt(3)/2 * geom.assemblyPitch*2
        geom.boundingBoxHeight = geom.assemblyPitch*2
        geom.axialSliceThickness = 1.0 # cm
        geom.longDiagonalOuter = 2*geom.assemblyPitch/np.sqrt(3) # point-to-point, outer duct (for specifying tally square)
        return geom

    def createPinCellUniverse(self,fuelOuterRadius:float, cladInnerRadius: float, cladOuterRadius: float, axialSliceThickness: float, 
        fuelMaterial: openmc.Material) -> openmc.Universe:
        '''
         Changes
        ----------
        01 Dec 2023 Andy - boundary type on top and bottom surfaces are implicit transport       
        05 Dec 2023 Andy - accept specific geometry components instead of geom CardContainer, for flexibility
        '''
        
        fuel_or = openmc.ZCylinder(r=fuelOuterRadius )
        clad_ir = openmc.ZCylinder(r=cladInnerRadius)
        clad_or = openmc.ZCylinder(r=cladOuterRadius)

        top = openmc.ZPlane(z0=+axialSliceThickness/2)
        bottom = openmc.ZPlane(z0=-axialSliceThickness/2)

        fuel_region = -fuel_or & -top & +bottom
        clad_region = +clad_ir & -clad_or  & -top & +bottom
        moderator_region = +clad_or & -top & +bottom

        clad_cell = openmc.Cell(fill=self.materials.getByID(self.claddingMaterialID), region=clad_region)
        sodium_cell = openmc.Cell(fill=self.materials.getByID(self.coolantMaterialID), region=moderator_region)


        inner_fuel_cell = openmc.Cell(fill=fuelMaterial, region=fuel_region)
        inner_u = openmc.Universe(cells=(inner_fuel_cell, clad_cell, sodium_cell))
        return inner_u


    def createAssemblyUniverse(self, sampleData:pd.DataFrame, geom: CardContainer) -> openmc.Universe:
        '''
         Changes
        ----------
        01 Dec 2023 Andy - boundary type on duct surface is implicit transport       
        '''
        # Create filling for emtpy space in the core
        sodium_mod_cell = openmc.Cell(fill=self.materials.getByID(self.coolantMaterialID))
        sodium_mod_u = openmc.Universe(cells=(sodium_mod_cell,))
        # Define a lattice for inner assemblies
        in_lat = openmc.HexLattice(name='inner assembly')
        in_lat.center = (0., 0.)
        in_lat.pitch = (sampleData['pin_pitch'],)
        in_lat.orientation = 'x'
        in_lat.outer = sodium_mod_u
        self.fillAssemblyLatticeWithPins(sampleData,geom,in_lat)
        # Create the prism that will contain the lattice
        outer_in_surface = openmc.model.hexagonal_prism(edge_length=geom.assemblySideLength, orientation='x')
        duct_surface = openmc.model.hexagonal_prism(edge_length=geom.assemblySideLength+sampleData['duct_thickness']*2/np.sqrt(3), 
            orientation='x')
        # Fill a cell with the lattice. This cell is filled with the lattice and contained within the prism.
        main_in_assembly = openmc.Cell(fill=in_lat, region= outer_in_surface)# & -top & +bottom)
        ducting=openmc.Cell(fill=self.materials.getByID(self.claddingMaterialID),region=~outer_in_surface & duct_surface)
        # Create a universe that contains both
        oneAssembly = openmc.Universe(cells=[main_in_assembly,ducting],name='one_assembly')#, out_in_assembly])
        return oneAssembly

    def fillAssemblyLatticeWithPins(self, sampleData:pd.DataFrame, geom:CardContainer, assemblyLattice:openmc.HexLattice) -> None:
        '''
        Override this method in a subclass to customize insertion of fuel pins into assembly.
        '''
        # Create rings of fuel universes that will fill the lattice
        pins_in_row = []
        for i in reversed(range(int(sampleData['num_rows']))):
            pins_in_row.append(i*6)
        pins_in_row[-1] = 1 # override loop for exception: 0th row has 1 pin.
        pinCellUniverse = self.createPinCellUniverse(geom.fuelOuterRadius,geom.cladInnerRadius,geom.cladOuterRadius, geom.axialSliceThickness,
            self.materials.getByID(self.fuelMaterialID))   
        assemblyLattice.universes = [[pinCellUniverse]*Npins for Npins in pins_in_row] 

    def createFullUniverse(self, sampleData:pd.DataFrame, geom: CardContainer) -> openmc.Universe:
        '''
         Changes
        ----------
        01 Dec 2023 Andy    - minor refactoring
        06 Dec 2023 Andy    - set coolant temp here      
        '''
        the_big_lattice=openmc.HexLattice()
        the_big_lattice.center = (0., 0.)
        the_big_lattice.pitch = (geom.assemblyPitch,)
        the_big_lattice.orientation = 'y'
        # Create filling for emtpy space outside the core
        coolantMaterial = self.materials.getByID(self.coolantMaterialID)
        coolantMaterial.temperature = sampleData['coolant_temperature']
        sodium_mod_cell = openmc.Cell(fill=coolantMaterial)
        sodium_mod_u = openmc.Universe(cells=(sodium_mod_cell,))
        the_big_lattice.outer = sodium_mod_u
        assemblyUniverse = self.createAssemblyUniverse(sampleData, geom) 
        layer1=[assemblyUniverse]
        layer2=[assemblyUniverse]*6
        #layer3=[assemblyUniverse]*12
        the_big_lattice.universes=[layer2,layer1]
        box = openmc.model.rectangular_prism(width=geom.boundingBoxWidth,height=geom.boundingBoxHeight,boundary_type='periodic')
        top = openmc.ZPlane(z0=+geom.axialSliceThickness/2, boundary_type='reflective')
        bottom = openmc.ZPlane(z0=-geom.axialSliceThickness/2, boundary_type='reflective')
        fullUniverseCell = openmc.Cell(fill=the_big_lattice,region=box & -top & +bottom)
        fullUniverse=openmc.Universe(cells=[fullUniverseCell],name='full_universe')#,outFullUniverseCell])
        return fullUniverse


    def settings(self,geom: CardContainer) -> openmc.Settings:
        '''
         Changes
        ----------
        01 Dec 2023 Andy - minor refactoring, adjust source sampling domain to coincide with geometry better.   
        '''        
        theSettings = openmc.Settings()
        theSettings.batches = 150
        theSettings.inactive = 50
        theSettings.particles = 1000
        theSettings.source = openmc.Source(space=openmc.stats.Box(
            (-geom.boundingBoxWidth/2, -geom.boundingBoxHeight/2, -0.5),
            (geom.boundingBoxWidth/2, geom.boundingBoxHeight/2, 0.5),
            only_fissionable=True
        ))
        return theSettings

    def tallies(self,geom: CardContainer) -> openmc.RegularMesh:
        Ngroups = 8
        resolution = [64,64]
        pitch = geom.longDiagonalOuter/2
        name = 'lo_fi'
        tally = src.tallies.meshTally(name,pitch,Ngroups,resolution)
        return [tally]

class HeterogeneousGeometryAssemblyModel(AssemblyModelSFR):
    '''
    This class only really exists to separate out functionality. Originally, the methods enabling generation of assemblies with heterogeneous pin dimensions
    were in AssemblyModelEBRIIDepletedFuel, which also requires materials to be heterogeneous. However in some cases one might want homogeneous materials in
    heterogeneous pin geometries, hence the separation of this functionality into this class.
    '''

    def fillAssemblyLatticeWithPins(self, sampleData:pd.DataFrame, geom:CardContainer, assemblyLattice:openmc.HexLattice) -> None:
        # Create rings of fuel universes that will fill the lattice
        pins_in_row = []
        pin_indices_in_row = []
        totalPinsInAssembly = ut.total_pins_hex(sampleData['num_rows'])
        topPinIdx = totalPinsInAssembly
        for i in reversed(range(int(sampleData['num_rows']))):
            pins_in_row.append(i*6)
            pin_indices_in_row.append(np.arange(topPinIdx-pins_in_row[-1],topPinIdx))
            if i==0:
                continue
            topPinIdx = pin_indices_in_row[-1][0]
        pins_in_row[-1] = 1 # override loop for exception: 0th row has 1 pin.
        pin_indices_in_row[-1] = [0]
        assemblyLattice.universes = [self.sampleFuelPinsCellUniversesRow(pinIndices,geom,sampleData) for pinIndices in pin_indices_in_row]

    def sampleFuelPinsCellUniversesRow(self, pinIndices:np.array, geom:CardContainer, sampleData:pd.DataFrame) -> list:
        fuelTypes = sampleData['fuel_type']
        burnupSteps = sampleData['burnup_step']
        pinTemperatures = sampleData['pin_temperature']
        overrides = sampleData['mask_overrides']
        pinCellUniverses = [self.createPinCellUniverse(geom.fuelOuterRadius[idx], geom.cladInnerRadius[idx], 
            geom.cladOuterRadius[idx], geom.axialSliceThickness, 
            self.selectPinMaterial(fuelTypes[idx],burnupSteps[idx],pinTemperatures[idx],overrides[idx])) for idx in pinIndices]
        return pinCellUniverses

    def selectPinMaterial(self, fuelType:int, burnupStep:int, pinTemperature:float=500.0,override=None) -> openmc.Material:
        '''
        Method signature is for polymorphism with AssemblyModelEBRIIDepletedFuel.selectPinMaterial()
        '''
        fuelMat = self.getMaterialByID(self.fuelMaterialID)
        fuelMat.temperature = pinTemperature
        return fuelMat

class AssemblyModelEBRIIDepletedFuel(HeterogeneousGeometryAssemblyModel):

    def __init__(self, depletedMaterialsDirectories: list, burnableMatID:int=0):
        super().__init__()
        self.populateDepletedMaterials(depletedMaterialsDirectories, burnableMatID)
        
    def populateMaterials(self):
        self.claddingMaterialID = self.materials.append(FluxUpMaterials.claddingHT9())
        self.coolantMaterialID  = self.materials.append(FluxUpMaterials.coolantNa())
        self.absorberMaterialID = self.materials.append(FluxUpMaterials.absorberEBRII())
    
    def getFuelMaterials(self) -> list:
        return [self.materials.getByID(ID) for ID in np.array(self.fuelMaterialIDs).ravel()]

    def disableDepletion(self):
        for fuelMat in self.getFuelMaterials():
            fuelMat.depletable = False
        absMat = self.getAbsorberMaterial()
        absMat.depletable = False
        
    def selectPinMaterial(self, fuelType:int, burnupStep:int, pinTemperature:float=500.0,override=None) -> openmc.Material:
        if override == None:
            fuelID = self.fuelMaterialIDs[fuelType][burnupStep]
            fuelMat = self.materials.getByID(fuelID)
            fuelMat.temperature = pinTemperature # NB if the same fuel material is requested twice in an assembly, all fuel pins of the same material will be at the same temperature
            return fuelMat
        if override=='B4C':
            absorberMat = self.materials.getByID(self.absorberMaterialID)
            absorberMat.temperature = pinTemperature
            return absorberMat
        if override=='Empty':
            return self.materials.getByID(self.coolantMaterialID)
