from openmc.deplete import results
from openmc import Materials
from openmc.exceptions import DataError
from openmc.data.library import DataLibrary
import openmc.checkvalue as cv
from src.FluxUpMaterials import FluxUpMaterials

class FluxUpDepletionResults(results.Results):
    
    def __init__(self,filepath=None,h5Filename='depletion_results.h5',materialsXmlFilename='materials.xml',crossSectionsXmlFullPath=None):
        h5FileFullPath = filepath + h5Filename
        self.filepath = filepath
        self.materialsXmlFilename = materialsXmlFilename
        self.crossSectionsXmlFullPath = crossSectionsXmlFullPath
        super().__init__(h5FileFullPath)


    


    ''' Override super to enable custom path for materials.xml file'''
    def export_to_materials(self, burnup_index, nuc_with_data=None) -> FluxUpMaterials:
        """Return openmc.Materials object based on results at a given step

        .. versionadded:: 0.12.1

        Parameters
        ----------
        burn_index : int
            Index of burnup step to evaluate. See also: get_step_where for
            obtaining burnup step indices from other data such as the time.
        nuc_with_data : Iterable of str, optional
            Nuclides to include in resulting materials.
            This can be specified if not all nuclides appearing in
            depletion results have associated neutron cross sections, and
            as such cannot be used in subsequent transport calculations.
            If not provided, nuclides from the cross_sections element of
            materials.xml will be used. If that element is not present,
            nuclides from OPENMC_CROSS_SECTIONS will be used.

        Returns
        -------
        mat_file : Materials
            A modified Materials instance containing depleted material data
            and original isotopic compositions of non-depletable materials
        """
        result = self[burnup_index]

        # Only materials found in the original materials.xml file will be
        # updated. If for some reason you have modified OpenMC to produce
        # new materials as depletion takes place, this method will not
        # work as expected and leave out that material.
        mat_file = FluxUpMaterials.from_xml(self.filepath + self.materialsXmlFilename)

        # Only nuclides with valid transport data will be written to
        # the new materials XML file. The precedence of nuclides to select
        # is first ones provided as a kwarg here, then ones specified
        # in the materials.xml file if provided, then finally from
        # openmc.config['cross_sections'].
        if nuc_with_data:
            cv.check_iterable_type('nuclide names', nuc_with_data, str)
            available_cross_sections = nuc_with_data
        else:
            # select cross_sections.xml file to use
            if mat_file.cross_sections:
                this_library = DataLibrary.from_xml(path=mat_file.cross_sections)
            else:
                this_library = DataLibrary.from_xml()

            # Find neutron libraries we have access to
            available_cross_sections = set()
            for lib in this_library.libraries:
                if lib['type'] == 'neutron':
                    available_cross_sections.update(lib['materials'])
            if not available_cross_sections:
                raise DataError('No neutron libraries found in cross_sections.xml')

        # Overwrite material definitions, if they can be found in the depletion
        # results, and save them to the new depleted xml file.
        for mat in mat_file:
            mat_id = str(mat.id)
            if mat_id in result.index_mat:
                mat.volume = result.volume[mat_id]

                # Change density of all nuclides in material to atom/b-cm
                atoms_per_barn_cm = mat.get_nuclide_atom_densities()
                for nuc, value in atoms_per_barn_cm.items():
                    mat.remove_nuclide(nuc)
                    mat.add_nuclide(nuc, value)
                mat.set_density('sum')

                # For nuclides in chain that have cross sections, replace
                # density in original material with new density from results
                for nuc in result.index_nuc:
                    if nuc not in available_cross_sections:
                        continue
                    atoms = result[0, mat_id, nuc]
                    if atoms > 0.0:
                        atoms_per_barn_cm = 1e-24 * atoms / mat.volume
                        mat.remove_nuclide(nuc) # Replace if it's there
                        mat.add_nuclide(nuc, atoms_per_barn_cm)

        return mat_file
   