from openmc import Material, Materials
from openmc.exceptions import InvalidIDError
import numpy as np


'''
REFERENCES:
[1] https://www.sciencedirect.com/science/article/pii/S0029549316304472, https://github.com/mit-crpg/openmc-reactor-examples/blob/main/SFR.ipynb
[2] Waltar & Todd
[3] Compendium of Material Composition Data for Radiation Transport Modeling PNNL 2006 https://www.pnnl.gov/main/publications/external/technical_reports/PNNL-15870.pdf
[4] BEAVRS spec
[5] Structural Materials for Liquid Metal Cooled Fast Reactor Fuel Assemblies, IAEA 1548
'''

class FluxUpMaterials(Materials):
    '''    Changes
    ----------
    01 Dec 2023 Andy    - Add more type hinting.
                        - New method fuelUO2_9ao.
    '''

    def allIDs(self) -> list:
        return [mat.id for mat in self]

    def getByID(self,idNumber:int) -> Material:
        for i in range(0,len(self)):
            if self[i].id == idNumber:
                return self[i]
        raise InvalidIDError('Provided idString does not exist.')

    def appendFromDepletionResults(self,aFluxUpDepletionResults,burnableMatId=1) -> list:
        '''
        Append a depleted material. All of the burnup steps of the depleted material are appended.
        In a depletion result, each burnup step of the material has the same ID. This method updates
        the material ID such that all depleted materials are regarded as independent, and stored just
        like any other material.

        Parameters
        ----------
            aFluxUpDepletionResults: An initialized instance of FluxUpDepletionResults
            burnableMatId: int
                An integer indicating the ID of the material in aFluxUpDepletionResults that should be added.

        Returns
        ----------
            newIDs: list
                A list of the IDs corresponding to the materials that were appended.

        Notes:
        ----------
            See the materials.xml input file for an OpenMC depletion simulation to find the correct burnable material ID.
            The class FluxUpDepletionResults loads all materials in the materials.xml, but only updates the materials with ID
            given in the depletion results h5 file (whose default name is depletion_results.h5).
        '''
        newIDs = []
        NburnSteps = len(aFluxUpDepletionResults)
        for i in range(0,NburnSteps):
            materialsi = aFluxUpDepletionResults.export_to_materials(i)
            burnableMateriali = materialsi.getByID(burnableMatId)
            burnableMateriali.id = None # invokes setter method in a superclass that automatically assigns an unused ID.
            newIDs.append(burnableMateriali.id)
            self.append(burnableMateriali)
        return newIDs

    
    def append(self,aMaterial:Material) -> int:
        super().append(aMaterial)
        return aMaterial.id

    def appendESFRMaterials(self) -> list:
        '''
        Append materials relevant to an SFR (European SFR in this case).
        Materials are very generic: Sodium, "fuel": fresh UPuO, cladding (Steel) etc
        Changes 
        ----------
        01 Dec 2023 Andy - minor edits
        '''
        # fuel density, 80U20Pu from [2] page 319 table 11.3 
        fuel = FluxUpMaterials.fuelESFR()
        #fuel = FluxUpMaterials.fuelUO2_9ao()
        fuel.temperature = 1500 # [x]
        smearDensity = 0.75 # [x]
        fuel.set_density('g/cm3', 11.08*smearDensity ) # 11.08 from [2] - unclear if it's smear-corrected
        coolant = FluxUpMaterials.coolantNa()
        coolant.temperature = 700 # [x]                             
        clad = FluxUpMaterials.claddingHT9()
        clad.temperature = 900 # [x]
        return [self.append(mat) for mat in [fuel,coolant,clad]]


    @staticmethod
    def fuelESFR() -> Material:
        '''
        MOX fuel with isotopics from European SFR, described in [1].
        Changes:
        ----------
        01 Dec 2023 Andy - Specify default density (can be overridden).
        21 Dec 2023 Andy - Corrected to be consistent with [1].
        22 Dec 2023 Andy - Correct fuel quantity to wo duh
        '''
        # Pu isotopics from [1], Outer fuel zone
        fuel = Material(name='fuel')
        fuel.add_components(components={'U':{'percent':73.00,
                                             'enrichment':0.18},
                                        'Pu238':0.54,     
                                        'Pu239':7.11,
                                        'Pu240':4.45,
                                        'Pu241':1.24,
                                        'Pu242':1.56,
                                        'Am241':0.17,
                                        'O':11.75},
                            percent_type='wo')
        fuel.temperature = 1500 # [1]

        
        return fuel

    @staticmethod
    def claddingHT9() -> Material:
        '''
        [3] page 111 (pdf), 102 (paper)
        '''
        steel = Material(name='Stainless Steel')
        steel.set_density('g/cm3', 7.894)
        steel.add_element('C', 100*0.002, 'wo')
        steel.add_element('Si', 100*0.004, 'wo')
        steel.add_element('P', 100*0.0003, 'wo')
        steel.add_element('S', 100*0.0002, 'wo')
        steel.add_element('V', 100*0.003, 'wo')
        steel.add_element('Cr', 100*0.115, 'wo')
        steel.add_element('Mn', 100*0.006, 'wo')
        steel.add_element('Fe', 100*0.8495, 'wo')
        steel.add_element('Ni',.100*0.005,'wo')
        steel.add_element('Mo',100*0.01,'wo')
        steel.add_element('W',100*0.005,'wo')
        steel.temperature = 900
        return steel

    @staticmethod
    def claddingODS2Y1() -> Material:
        '''
        [5] p34 (Table 7)
        '''
        elements = ['C', 'Mn', 'P', 'S', 'Si', 'Ni', 'Cr', 'Mo', 'V', 'Ti', 'Co', 'Cu', 'Al', 'B', 'W', 'Zr', 'N', 'O', 'Y']
        percent = np.array([0.045, 0.04, 0.001, 0.002, 0.03, 0.24, 12.85, 0.03, 0.007, 0.003, 0.005, 0.01, 0.007, 0.004, 0.01, 0.003, 0.017, 0.15, 0.20])
        NFe = 100.0 - np.sum(percent)
        ods = Material(name='ODS Steel')
        ods.set_density('g/cm3', 7.894)
        [ods.add_element(elements[i], percent[i],'ao') for i in range(0,len(elements))]
        ods.add_element('Fe',NFe,'ao')
        return ods
        


    @staticmethod
    def coolantNa() -> Material:
        coolant = Material(name='coolant')
        coolant.add_nuclide('Na23', 100.0, 'ao') # [x]
        coolant.set_density('g/cm3', 0.96) # [x]
        return coolant

    @staticmethod
    def coolantH2O() -> Material:
        coolant = Material(name='coolant')
        coolant.add_nuclide('H1', 66.666, 'ao') # [x]
        coolant.add_nuclide('O16', 33.333, 'ao') # [x]
        coolant.add_s_alpha_beta('c_H_in_H2O')

        coolant.set_density('g/cm3', 0.740582) # [x]
        return coolant

    @staticmethod
    def absorberEBRII() -> Material:
        '''
        [4] page 47
        '''
        absorber = Material(name='absorber')
        absorber.add_nuclide('B10', 1.5203e-2, 'ao')
        absorber.add_nuclide('B11', 6.1504e-2, 'ao')
        absorber.add_nuclide('C12', 1.8972e-2, 'ao')
        absorber.add_nuclide('C13', 2.1253e-2, 'ao')
        absorber.set_density('g/cm3', 1.76)
        return absorber


    @staticmethod
    def fuelUO2_9ao() -> Material:
        '''
        UO2, 9% enriched
        '''
        fuel = Material(name='fuel')
        fuel.add_components(components={'U':{'percent':1/3*100,
                                             'enrichment':9.0},
                                        'O':2/3*100},
                            percent_type='ao')
        fuel.set_density('g/cm3',10.29769)
        return fuel

    @staticmethod
    def fuelUO2(enrichment:float=5.0) -> Material:
        fuel = Material(name='fuel')
        fuel.add_components(components={'U':{'percent':1/3*100,
                                             'enrichment':enrichment},
                                        'O':2/3*100},
                            percent_type='ao')
        fuel.set_density('g/cm3',10.29769)
        return fuel

    @staticmethod
    def fuelMOX(Pu:Material,Uwo:float=80.0,Uenrich:float=5.0):
        u = Material(name='U')
        u.add_element('U',1.0,enrichment=Uenrich)
        u.set_density('g/cm3',18.85)
        Puwo = 100.0 - Uwo
        HM = Material.mix_materials([u,Pu],[1e-2*Uwo,1e-2*Puwo],'wo')
        O = Material(name='Oxygen')
        O.add_element('O',1.0)
        fuel = Material.mix_materials([HM,O],[1/3,2/3],'ao')
        fuel.name = 'fuel'
        fuel.set_density('g/cm3',10.29769)
        return fuel

    @staticmethod
    def fuelPureU(Uenrich:float=5.0):
        u = Material(name='U')
        u.add_element('U',1.0,enrichment=Uenrich)
        u.set_density('g/cm3',18.85)
        return u


    @staticmethod
    def fuelUPuZr(Pu:Material,Uwo:float=80.0,Uenrich:float=5.0,Zrwo:float=10.0,smearDensity:float=75.0) -> Material:
        '''
        Parameters
        ----------
            Pu: openmc.Material

            Uwo: float
                weight % of U [0-100]
            Uenrich: float
                U enrichment [0-100]
            Zrwo: float
                weight % of Zr [0-100]. NB increasing fraction of Zr above 0 causes proportional reduction in Pu and U masses.
            smearDensity: float
                smear density in % of the fuel [0-100].
        Returns
        ----------
            cont: CardContainer
                Instance of a CardContainer holding onto geometric quantities
        Notes:
        ----------
        E.g. Let's define a fuel with 20w/o Zr and UPu in the balancing 80w/o. The HM component is 90 w/o U and 10 w/o Pu. The resulting
        mixture would be by weight 20 w/o Zr, 72 w/o U and 8 w/o Pu.
        Defaults from Waltar & Todd, page 319 for example.

        Changes
        '''
        u = Material(name='U')
        u.add_element('U',1.0,enrichment=Uenrich)
        u.set_density('g/cm3',18.85)
        Zr = Material(name='Zr')
        Zr.add_element('Zr',1.0)
        Zr.set_density('g/cm3',6.52)
        HMwo = 100.0 - Zrwo
        Puwo = 100.0 - Uwo
        fuel = Material.mix_materials([u,Pu,Zr],[1e-4*Uwo*HMwo,1e-4*Puwo*HMwo,0.01*Zrwo],'wo')
        #fuel = Material.mix_materials([u,Pu,Zr],[1e-2*Uwo*HMwo,1e-2*Puwo*HMwo,Zrwo],'wo')
        fuel.name = 'fuel'
        fuel.set_density('g/cm3',fuel.density*0.01*smearDensity)
        return fuel

    @staticmethod
    def puVectorFromDepletionResult(dir:str,buStep:int=1,PuIsotopeNames:list = ['Pu236','Pu237','Pu238','Pu239','Pu240','Pu241','Pu242','Pu243']):
        from src.FluxUpDepletionResults import FluxUpDepletionResults
        fudr = FluxUpDepletionResults(dir, crossSectionsXmlFullPath=Materials.cross_sections)
        fum = fudr.export_to_materials(buStep,PuIsotopeNames)
        PuIsotopeao = np.array([fum.getByID(1).get_nuclide_densities()[PuIsoName][1] for PuIsoName in PuIsotopeNames])
        PuIsotopeao = 100.0*np.divide(PuIsotopeao, np.sum(PuIsotopeao))
        puMaterial = Material(name='Pu')
        [puMaterial.add_nuclide(PuIsotopeNames[i],PuIsotopeao[i],'ao') for i in range(len(PuIsotopeNames))]
        puMaterial.set_density('g/cm3',19.84)
        return puMaterial
