import numpy as np
from src.assembly_model import assembly_model
from src.tallies import meshTally
import openmc
import os,glob
import h5py
import pandas as pd

class OpenMCWrapper:
    '''
    Changes
    ----------
    01 Dec 2023 Andy - Refactoring for better inheritance.
    '''
    def __init__(self, casesDataFrame:pd.DataFrame, NgroupsIn=15,NgroupsOut=30,resolutionIn=[32,32], resolutionOut=[128,128]):
        self.NgroupsIn = NgroupsIn
        self.NgroupsOut = NgroupsOut
        self.resolutionIn = resolutionIn
        self.resolutionOut = resolutionOut
        self.casesDataFrame = casesDataFrame
        self.openMCExecutable = '/usr/bin/openmc' # user can set externally. Default is /usr/bin/openmc
        self.loFiResultsArray = np.zeros(shape=(len(self.casesDataFrame),NgroupsIn,resolutionIn[0],resolutionIn[1]))
        self.hiFiResultsArray = np.zeros(shape=(len(self.casesDataFrame),NgroupsOut,resolutionOut[0],resolutionOut[1]))
        
        self.loFiRelErrArray = np.zeros(shape=(len(self.casesDataFrame),NgroupsIn,resolutionIn[0],resolutionIn[1]))
        self.hiFiRelErArray = np.zeros(shape=(len(self.casesDataFrame),NgroupsOut,resolutionOut[0],resolutionOut[1]))

        self.loFiStdDevArray = np.zeros(shape=(len(self.casesDataFrame),NgroupsIn,resolutionIn[0],resolutionIn[1]))
        self.hiFiStdDevArray = np.zeros(shape=(len(self.casesDataFrame),NgroupsOut,resolutionOut[0],resolutionOut[1]))

        self.loFiName = 'lo_fi'
        self.hiFiName = 'hi_fi'
        self.relErrStr = 'rel_err'
        self.stdDevStr = 'std_dev'
        self.saveUncertainty = False
        self.debugMode = True
        # these instance variables hold on to lists of functions that are called sequentially to generate a simulation model, and do model
        # overrides. Order is important.
        self.preRunModelOverrides = [] # should be a list of handles to functions that modify the model immediately before running OpenMC.
        self.customModelFunctions = [assembly_model] # should be a list of handles to functions that generate the simulation model (assembly or whatever)
        self.customPlotColorFunctions = []
        self.customData = {} # used to pass custom data to custom model functions if needed. Default is an empty Dict.

    def createPlotsOnly(self):
        for i in range(0,len(self.casesDataFrame)):
            model = self.generateModel(i)
            model.export_to_xml()
            self.createPlots(model,i)

    def runAllOpenMC(self):
        for i in range(0,len(self.casesDataFrame)):
            model = self.generateModel(i)
            loFiTally = self.generateTally(model, self.loFiName, self.NgroupsIn,self.resolutionIn)
            hiFiTally = self.generateTally(model, self.hiFiName, self.NgroupsOut,self.resolutionOut)
            model.tallies = [loFiTally,hiFiTally]
            model = self.applyPreRunModelOverrides(model)
            lastStatepoint = model.run(openmc_exec=self.openMCExecutable,output=True)
            self.createPlots(model,i) if self.debugMode else print("Debug mode off, no geometry plot output")
            datai = self.postProcessSingleRun(lastStatepoint)
            self.appendResultToResults(i,datai)

        
    def generateModel(self,idx:int):
        sampleData = self.casesDataFrame.iloc[idx]
        for customFunction in self.customModelFunctions:
            model = customFunction(sampleData,self.customData)
        return model

    def generateTally(self, model, name,Ngroups,resolution):
        meshTallyWidth = self.customData['mesh_tally_width'] if 'mesh_tally_width' in self.customData else model.assemblyPitch
        tally = meshTally(name,meshTallyWidth,Ngroups,resolution)
        return tally

    def postProcessSingleRun(self,lastStatepoint:str) -> dict:
        '''
        Postprocessing after each run: reshaping tallies into convenient shape, collect all tallies in a container.

        Parameters
        ----------
            latStatePoint: 
                Path to last statepoint file.
        Returns
        ----------
            data: dict
                A dictionary of tally results

        Notes:
        ----------
        Changes:
        ----------
        01 Dec 2023 Andy - Correct reshape order: rows and columns were reversed but it hadn't mattered previously as Nx was = Ny.
        '''
        with openmc.StatePoint(lastStatepoint) as sp:
            loFiTally = sp.get_tally(name=self.loFiName)
            hiFiTally = sp.get_tally(name=self.hiFiName)
            #energyFilter = loFiTally.filters[0]
            #meshFilter = loFiTally.filters[1]
            loFiResult = loFiTally.get_values(value='mean').ravel()
            hiFiResult = hiFiTally.get_values(value='mean').ravel()
            loFiRelErr = loFiTally.get_values(value=self.relErrStr).ravel()
            hiFiRelErr = hiFiTally.get_values(value=self.relErrStr).ravel()
            loFiStd = loFiTally.get_values(value=self.stdDevStr).ravel()
            hiFiStd = hiFiTally.get_values(value=self.stdDevStr).ravel()
            # NB loFiResult3D(0,0,0) is (1st energy group, 1st y-coord, 1st x-coord)
            # the 1st x/y coordinate refers to the lower left corner of the assembly
            # even though in memory it's usual matrix row-column notation from top down. 
            loFiResult3D = np.reshape(loFiResult,(self.NgroupsIn,self.resolutionIn[1],self.resolutionIn[0]))
            hiFiResult3D = np.reshape(hiFiResult,(self.NgroupsOut,self.resolutionOut[1],self.resolutionOut[0]))
            loFiRelErr3D = np.reshape(loFiRelErr,(self.NgroupsIn,self.resolutionIn[1],self.resolutionIn[0]))
            hiFiRelErr3D = np.reshape(hiFiRelErr,(self.NgroupsOut,self.resolutionOut[1],self.resolutionOut[0]))
            loFiStd3D = np.reshape(loFiStd,(self.NgroupsIn,self.resolutionIn[1],self.resolutionIn[0]))
            hiFiStd3D = np.reshape(hiFiStd,(self.NgroupsOut,self.resolutionOut[1],self.resolutionOut[0]))

            data = {self.loFiName:loFiResult3D,self.hiFiName:hiFiResult3D,
                self.loFiName + '_' + self.relErrStr:loFiRelErr3D, self.hiFiName + '_' + self.relErrStr:hiFiRelErr3D,
                self.loFiName + '_' + self.stdDevStr:loFiStd3D, self.hiFiName + '_' + self.stdDevStr:hiFiStd3D}
        return data

    def deleteFiles(self):
        fileList = ["geometry.xml","materials.xml","settings.xml","tallies.out","tallies.xml","summary.h5","plots.xml"]
        for file in fileList:
            os.remove(os.path.join(os.getcwd(), file))   
        statePointFiles = glob.glob(os.path.join(os.getcwd(), "statepoint.*"))
        for file in statePointFiles:
            os.remove(file)


    def saveResults(self,hdf5FileName:str):
        with h5py.File(hdf5FileName, 'w') as f:
            dataset = f.create_dataset(self.loFiName, data=self.loFiResultsArray)
            dataset = f.create_dataset(self.hiFiName, data=self.hiFiResultsArray)
            if(self.saveUncertainty):
                dataset = f.create_dataset(self.loFiName + '_' + self.relErrStr, data=self.loFiRelErrArray)
                dataset = f.create_dataset(self.hiFiName + '_' + self.relErrStr, data=self.hiFiRelErArray)
                dataset = f.create_dataset(self.loFiName + '_' + self.stdDevStr, data=self.loFiStdDevArray)
                dataset = f.create_dataset(self.hiFiName + '_' + self.stdDevStr, data=self.hiFiStdDevArray)
        # save input data (DataFrame) by appending
        self.casesDataFrame.to_hdf(hdf5FileName,key='input')

    def createPlots(self, model:openmc.model.Model, idx=0):
        print("Creating plots...")
        plot = openmc.Plot()
        plot.filename = 'geometry_plot_{:d}'.format(idx)
        plot.width = (model.assemblyPitch, model.assemblyPitch)
        plot.pixels = (800,800)
        plot.color_by = 'material'
        
        if len(self.customPlotColorFunctions) == 0:
            plot.colors = {model.beavrsMaterials['Borated Water']: 'cyan',
                model.beavrsMaterials['B4C']: 'gray',
                model.hot_water: 'blue', 
                model.clad:'gray',
                model.beavrsMaterials['Fuel 1.6%']:'red',
                model.beavrsMaterials['Fuel 2.4%']:'red',
                model.beavrsMaterials['Fuel 3.2%']:'red',
                model.beavrsMaterials['Fuel 3.4%']:'red',
                model.beavrsMaterials['Fuel 3.1%']:'red'
                }
        for customFunction in self.customPlotColorFunctions:
            plot.colors = customFunction(self, model)

        #plot.colors = {model.fuel: 'red', model.hot_water: 'blue', model.clad:'gray'}
        plots = openmc.Plots([plot])
        plots.export_to_xml()
        openmc.plot_geometry(openmc_exec=self.openMCExecutable)

    def applyPreRunModelOverrides(self,model:openmc.model.Model) -> openmc.model.Model:
        for overrideFunction in self.preRunModelOverrides:
            model = overrideFunction(model)
        return model

    def appendResultToResults(self, idx:int, singleResult:dict) -> None:
        self.loFiResultsArray[idx,:,:,:] = singleResult[self.loFiName]
        self.hiFiResultsArray[idx,:,:,:] = singleResult[self.hiFiName]
        self.loFiRelErrArray[idx,:,:,:] = singleResult[self.loFiName + '_' + self.relErrStr]
        self.hiFiRelErArray[idx,:,:,:] = singleResult[self.hiFiName + '_' + self.relErrStr]
        self.loFiStdDevArray[idx,:,:,:] = singleResult[self.loFiName + '_' + self.stdDevStr]
        self.hiFiStdDevArray[idx,:,:,:] = singleResult[self.hiFiName + '_' + self.stdDevStr]