
import numpy as np
import tensorflow as tf
import os, sys
import matplotlib.pyplot as plt
homeDir = os.path.expanduser("~")
# Edit this path to point to where your working copy of the flux-up code is stored.
sys.path.insert(0,homeDir + "/codes/flux-up")
sys.path.insert(0,homeDir + "/.local/lib/python3.9/site-packages")
import src
import h5py
import src.normalization as norm
from abc import ABC as AbstractBaseClass, abstractmethod



class AbstractPipeline(AbstractBaseClass):
    pass

class BasicPipeline(AbstractPipeline):


    def connect(self, trainDataPathList: list, valDatapathList: list) -> tuple:
            output_signature = self.outputSignature()
            train_ds, val_ds = self.instantiateDataset(trainDataPathList, valDatapathList, output_signature)

            #print('Maximum training values pre-scaling: ', norm.spatial_max(iter(train_ds)))
            #print('Maximum validation values pre-scaling: ', norm.spatial_max(iter(val_ds)))
            train_ds, val_ds, normX, normY = self.scale(train_ds,val_ds)
            train_ds, val_ds = self.postScaling(train_ds,val_ds)

            AUTOTUNE = tf.data.AUTOTUNE
            train_ds = train_ds.cache().shuffle(10000).prefetch(buffer_size = AUTOTUNE)
            val_ds = val_ds.cache().prefetch(buffer_size = AUTOTUNE)
            return train_ds, output_signature, val_ds, normX, normY
    
    def postScaling(self,train_ds,val_ds):
        '''
        Subclasses should override for custom behavior
        '''
        return train_ds, val_ds

    def instantiateDataset(self,trainDataPathList, valDatapathList, output_signature):
        
        train_ds = tf.data.Dataset.from_generator(src.TallyGenerator(trainDataPathList,training=True),
                                                output_signature = output_signature)
        val_ds = tf.data.Dataset.from_generator(src.TallyGenerator(valDatapathList,training=False),
                                                output_signature = output_signature)
        return train_ds,val_ds

    def outputSignature(self) -> tuple:
        output_signature = (tf.TensorSpec(shape = (None, None, None, 8), dtype = tf.float32),
                        tf.TensorSpec(shape = (None, None, None, 16), dtype = tf.float32))
        return output_signature

    def volumeScaling(self,train_ds: tf.data.Dataset, val_ds: tf.data.Dataset) -> tuple:
        ''' Scale by "volume" of simulation '''
        train_ds = train_ds.map(lambda X,Y: (X*tf.cast(tf.shape(X)[1]*tf.shape(X)[2], tf.float32), 
                                            (Y*tf.cast(tf.shape(X)[1]*tf.shape(X)[2], tf.float32))))
        val_ds = val_ds.map(lambda X,Y: (X*tf.cast(tf.shape(X)[1]*tf.shape(X)[2], tf.float32), 
                                            (Y*tf.cast(tf.shape(X)[1]*tf.shape(X)[2], tf.float32))))
        return train_ds, val_ds
    
    def normalizeSpatialMax(self,train_ds: tf.data.Dataset, val_ds: tf.data.Dataset) -> tuple:
        ''' Normalize so data is approx. between 0 and 1'''
        normX, normY = norm.spatial_max(iter(train_ds))
        train_ds = train_ds.map(lambda X,Y: (tf.divide(X,normX),
                                            tf.divide(Y,normY)))
        val_ds = val_ds.map(lambda X,Y: (tf.divide(X,normX),
                                        tf.divide(Y,normY)))
        return train_ds,val_ds,normX,normY

    def scale(self,train_ds: tf.data.Dataset, val_ds: tf.data.Dataset) -> tuple:
        train_ds,val_ds = self.volumeScaling(train_ds,val_ds)
        train_ds,val_ds,normX,normY = self.normalizeSpatialMax(train_ds,val_ds)
        return train_ds,val_ds,normX,normY

    def predictForExport(self,model, dataset):
        data = [[sampleX.numpy(), model.predict(sampleX), sampleY.numpy()] for sampleX,sampleY in dataset]
        return data

    def export(self, val_ds: tf.data.Dataset, valDatapath: list ,output_signature: tuple, model:tf.keras.Model, outFileName:str = 'out.h5') -> None:
        data = self.predictForExport(model,val_ds)
        #data = [[sampleX.numpy(), model.predict(sampleX), sampleY.numpy()] for sampleX,sampleY in val_ds]
        X_test_scaled = [sample[0] for sample in data]
        y_scaled =  [sample[1] for sample in data]
        y_test_scaled = [sample[2] for sample in data]
        #%%
        val_uncertainty_ds = tf.data.Dataset.from_generator(src.TallyGenerator(valDatapath,training=False,tallyNames=('lo_fi_rel_err','hi_fi_rel_err')),
                                                output_signature = output_signature)
        data = [[result[0].numpy(),result[1].numpy()] for result in val_uncertainty_ds] # result 0 and 1 are expected to be given by order of above tallyNames
        loFiRelErr = [sample[0] for sample in data]
        hiFiRelErr = [sample[1] for sample in data]
        #%%
        import h5py
        Nsamples = len(y_scaled)
        with h5py.File(outFileName,'w') as f:
            for i in range(0,Nsamples):
                group = f.create_group('sample-' + str(i))
                group.create_dataset('X_test_scaled', data=X_test_scaled[i])
                group.create_dataset('y_test_scaled', data=y_test_scaled[i])
                group.create_dataset('y_scaled', data=y_scaled[i])
                group.create_dataset('loFiRelErr', data=loFiRelErr[i])
                group.create_dataset('hiFiRelErr', data=hiFiRelErr[i])

class RelativeErrorBasicPipeline(BasicPipeline):

    def __init__(self,tallyGeneratorClass=src.RelativeErrorTallyGenerator):
        super().__init__()
        self.tallyGeneratorClass = tallyGeneratorClass

    def instantiateDataset(self,trainDataPathList, valDatapathList, output_signature):    
        train_ds = tf.data.Dataset.from_generator(self.tallyGeneratorClass(trainDataPathList,training=True, tallyNames=('lo_fi','hi_fi','lo_fi_rel_err','hi_fi_rel_err')),
                                                output_signature = output_signature)
        val_ds = tf.data.Dataset.from_generator(self.tallyGeneratorClass(valDatapathList,training=False,tallyNames=('lo_fi','hi_fi','lo_fi_rel_err','hi_fi_rel_err')),
                                                output_signature = output_signature)
        return train_ds,val_ds

    def outputSignature(self):
        output_signature = (tf.TensorSpec(shape = (None, None, None, 8), dtype = tf.float32),
                        tf.TensorSpec(shape = (None, None, None, 16), dtype = tf.float32),
                        tf.TensorSpec(shape = (None, None, None, 8), dtype = tf.float32),
                        tf.TensorSpec(shape = (None, None, None, 16), dtype = tf.float32))
        return output_signature

    def volumeScaling(self,train_ds: tf.data.Dataset, val_ds: tf.data.Dataset) -> tuple:
        ''' Scale by "volume" of simulation '''
        train_ds = train_ds.map(lambda X,Y,XrelErr,YrelErr: (X*tf.cast(tf.shape(X)[1]*tf.shape(X)[2], tf.float32), 
                                            (Y*tf.cast(tf.shape(X)[1]*tf.shape(X)[2], tf.float32)),
                                            XrelErr,YrelErr))
        val_ds = val_ds.map(lambda X,Y,XrelErr,YrelErr: (X*tf.cast(tf.shape(X)[1]*tf.shape(X)[2], tf.float32), 
                                            (Y*tf.cast(tf.shape(X)[1]*tf.shape(X)[2], tf.float32)),
                                            XrelErr,YrelErr))
        return train_ds, val_ds

    def normalizeSpatialMax(self,train_ds: tf.data.Dataset, val_ds: tf.data.Dataset) -> tuple:
        ''' Normalize so data is approx. between 0 and 1'''
        normX, normY = norm.spatial_max(iter(train_ds))
        train_ds = train_ds.map(lambda X,Y,XrelErr,YrelErr: (tf.divide(X,normX),
                                            tf.divide(Y,normY),XrelErr,YrelErr))
        val_ds = val_ds.map(lambda X,Y,XrelErr,YrelErr: (tf.divide(X,normX),
                                        tf.divide(Y,normY),XrelErr,YrelErr))
        return train_ds,val_ds,normX,normY

    def predictForExport(self,model, dataset):
        data = [[X.numpy(), model.predict([X,YrelErr]),  Y.numpy(), XrelErr.numpy(), YrelErr.numpy()] for X,Y,XrelErr,YrelErr in dataset]
        return data

    def export(self, val_ds: tf.data.Dataset, valDatapath: list ,output_signature: tuple, model:tf.keras.Model, outFileName:str = 'out.h5') -> None:
        data = self.predictForExport(model,val_ds)
        #data = [[sampleX.numpy(), model.predict(sampleX), sampleY.numpy()] for sampleX,sampleY in val_ds]
        X_test_scaled = [sample[0] for sample in data]
        y_scaled =  [sample[1][:,:,:,0:16] for sample in data] # nasty but ok
        y_test_scaled = [sample[2] for sample in data]
        loFiRelErr = [sample[3] for sample in data]
        hiFiRelErr = [sample[4] for sample in data]
        #%%
        import h5py
        Nsamples = len(y_scaled)
        with h5py.File(outFileName,'w') as f:
            for i in range(0,Nsamples):
                group = f.create_group('sample-' + str(i))
                group.create_dataset('X_test_scaled', data=X_test_scaled[i])
                group.create_dataset('y_test_scaled', data=y_test_scaled[i])
                group.create_dataset('y_scaled', data=y_scaled[i])
                group.create_dataset('loFiRelErr', data=loFiRelErr[i])
                group.create_dataset('hiFiRelErr', data=hiFiRelErr[i])

class SplitRelativeErrorBasicPipeline(RelativeErrorBasicPipeline):
    
    def instantiateDataset(self,trainDataPathList, valDatapathList, output_signature):    
        train_ds = tf.data.Dataset.from_generator(src.splitTensorTallyGenerator(trainDataPathList,training=True, tallyNames=('lo_fi','hi_fi','lo_fi_rel_err','hi_fi_rel_err')),
                                                output_signature = output_signature)
        val_ds = tf.data.Dataset.from_generator(src.RelativeErrorTallyGenerator(valDatapathList,training=False,tallyNames=('lo_fi','hi_fi','lo_fi_rel_err','hi_fi_rel_err')),
                                                output_signature = output_signature)
        return train_ds,val_ds         

    def scale(self,train_ds: tf.data.Dataset, val_ds: tf.data.Dataset) -> tuple:
        '''train_ds gets scaled volumetrically by the splitTensorTallyGenerator'''
        val_ds = self.volumeScaling(val_ds)
        train_ds,val_ds,normX,normY = self.normalizeSpatialMax(train_ds,val_ds)
        return train_ds,val_ds,normX,normY

    def volumeScaling(self,ds:tf.data.Dataset):
        ''' Scale by "volume" of simulation '''
        ds = ds.map(lambda X,Y,XrelErr,YrelErr: (X*tf.cast(tf.shape(X)[1]*tf.shape(X)[2], tf.float32), 
                                            (Y*tf.cast(tf.shape(X)[1]*tf.shape(X)[2], tf.float32)),
                                            XrelErr,YrelErr))
        return ds



class DynamicNormalizationPipeline(BasicPipeline):
    
    def scale(self, train_ds: tf.data.Dataset, val_ds: tf.data.Dataset) -> tuple:
        train_ds,val_ds = self.volumeScaling(train_ds,val_ds)
        train_ds,val_ds = self.normalizeDynamic(train_ds,val_ds)
        normX = None
        normY = None
        return train_ds,val_ds,normX,normY

    def normalizeDynamic(self, train_ds: tf.data.Dataset, val_ds: tf.data.Dataset) -> tuple:
        def customScale(X,Y):
            maxX_duplicate = tf.zeros((16,))
            maxX = tf.reduce_max(X,axis=[0,1,2])
            #maxX = tf.reduce_max(X,axis=[0,1,2])
            #maxX_duplicate[::2] = maxX
            #maxX_duplicate[1::2] = maxX
            #maxX_duplicate = tf.convert_to_tensor(maxX_duplicate,dtype=tf.float32)
            maxX_duplicate = tf.tensor_scatter_nd_update(maxX_duplicate, [[1],[3],[5],[7],[9],[11],[13],[15]], maxX)
            maxX_duplicate = tf.tensor_scatter_nd_update(maxX_duplicate, [[0],[2],[4],[6],[8],[10],[12],[14]], maxX)
            x_scaled = tf.divide(X,maxX)
            y_scaled = tf.divide(Y,maxX_duplicate)
            return x_scaled,y_scaled
        return train_ds.map(customScale), val_ds.map(customScale)

    