from src import OpenMCWrapper, AssemblyModel
import openmc
from src import utilities as ut
import numpy as np
import h5py

class SFROpenMCWrapper(OpenMCWrapper):
    '''
    Changes
    ----------
    01 Dec 2023 Andy - Lots of upgrades...
    '''

    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)
        self.loFiResultsArray = []
        self.hiFiResultsArray = []
        self.loFiRelErrArray = []
        self.hiFiRelErArray = []
        self.loFiStdDevArray = []
        self.hiFiStdDevArray = []
        self.asmModelBuilder = AssemblyModel.AssemblyModelSFR()
       

    def generateModel(self,idx:int) -> openmc.model.Model:
        sampleData = self.casesDataFrame.iloc[idx]
        model,geom = self.asmModelBuilder.openmcModel(sampleData,self.customData)
        self.geom = geom
        defaultGridSpacingLoRes = 0.33
        self.meshGridSpacingLoRes = self.customData['mesh_tally_dx'] if 'mesh_tally_dx' in self.customData else defaultGridSpacingLoRes
        self.meshGridSpacingHiRes = self.meshGridSpacingLoRes/2
        self.resolutionIn = np.floor(np.divide([self.geom.boundingBoxWidth,self.geom.boundingBoxHeight],self.meshGridSpacingLoRes)).astype(int)
        self.resolutionOut = self.resolutionIn*2
        return model

    def createPlots(self, model, idx:int=0) -> None:
        print("Creating plots...")
        plot = openmc.Plot()
        plot.filename = 'geometry_plot_{:d}'.format(idx)
        plot.width = (self.geom.boundingBoxWidth, self.geom.boundingBoxHeight)
        plot.pixels = (800, int(800*self.geom.boundingBoxHeight/self.geom.boundingBoxWidth))
        plot.color_by = 'material'
        for customFunction in self.customPlotColorFunctions:
            plot.colors = customFunction(self, model)
        plots = openmc.Plots([plot])
        plots.export_to_xml()
        openmc.plot_geometry(openmc_exec=self.openMCExecutable)

    def generateTally(self, model:openmc.model.Model, name:str, Ngroups:int, resolution:list) -> openmc.Tally:
        mesh = openmc.RegularMesh()
        mesh.dimension = resolution #indices in each direction: integer
        mesh.lower_left = [-self.geom.boundingBoxWidth/2, -self.geom.boundingBoxHeight/2] 
        mesh.upper_right = [self.resolutionIn[0]*self.meshGridSpacingLoRes/2, self.resolutionIn[1]*self.meshGridSpacingLoRes/2] # should be same as [self.resolutionOut[0]*self.meshGridSpacingHiRes/2, self.resolutionOut[1]*self.meshGridSpacingHiRes/2] 
        groups = np.logspace(np.log10(100), np.log10(20.0e6),Ngroups+1) # see Energies paper sup. info.
        energyFilter = openmc.EnergyFilter(groups)
        #energyFilter = openmc.EnergyFilter([1.0e5, 20.0e6])
        tally = openmc.Tally(name=name)
        tally.filters = [energyFilter,openmc.MeshFilter(mesh)]
        tally.scores = ['flux']
        tally.estimator = 'tracklength'
        return tally

    def appendResultToResults(self, idx:int, singleResult:dict) -> None:
        self.loFiResultsArray.append(singleResult[self.loFiName])
        self.hiFiResultsArray.append(singleResult[self.hiFiName])
        self.loFiRelErrArray.append(singleResult[self.loFiName + '_' + self.relErrStr])
        self.hiFiRelErArray.append(singleResult[self.hiFiName + '_' + self.relErrStr])
        self.loFiStdDevArray.append(singleResult[self.loFiName + '_' + self.stdDevStr])
        self.hiFiStdDevArray.append(singleResult[self.hiFiName + '_' + self.stdDevStr])

    def saveResults(self,hdf5FileName:str):
        Nsamples = len(self.casesDataFrame)
        with h5py.File(hdf5FileName, 'w') as f:
            for i in range(0,Nsamples):
                group = f.create_group('sample-' + str(i))
                group.create_dataset(self.loFiName, data=self.loFiResultsArray[i])
                group.create_dataset(self.hiFiName, data=self.hiFiResultsArray[i])
                if(self.saveUncertainty):
                    group.create_dataset(self.loFiName + '_' + self.relErrStr, data=self.loFiRelErrArray[i])
                    group.create_dataset(self.hiFiName + '_' + self.relErrStr, data=self.hiFiRelErArray[i])
                    group.create_dataset(self.loFiName + '_' + self.stdDevStr, data=self.loFiStdDevArray[i])
                    group.create_dataset(self.hiFiName + '_' + self.stdDevStr, data=self.hiFiStdDevArray[i])
        # save input data (DataFrame) by appending
        self.casesDataFrame.to_hdf(hdf5FileName,key='input')