import h5py
import os
import numpy as np
import random
from src.utilities import moveaxis


'''
Based on FrameGenerator from https://github.com/tensorflow/docs/blob/master/site/en/tutorials/load_data/video.ipynb
'''

class TallyGenerator:
    def __init__(self,tallyPaths:list,training:bool=False,tallyNames=('lo_fi','hi_fi')):
        self.tallyPaths = [os.path.join(filePath,'result.hdf5') for filePath in tallyPaths] 
        self.training = training
        self.tallyNames = tallyNames
        files,samples = self.getFilesAndSamples()
        self.pairs = list(zip(files, samples))
        if self.training:
            random.shuffle(self.pairs)
        

    def __call__(self):
        for file, sample in self.pairs:
            data = h5py.File(file)
            tallyX = moveaxis([data[sample][self.tallyNames[0]][:]])
            tallyY = moveaxis([data[sample][self.tallyNames[1]][:]])
            yield tallyX, tallyY
        

    def getFilesAndSamples(self) -> tuple:
        
        fileNamesList = []
        samplesList = []
        for tallyFile in self.tallyPaths:
            data = h5py.File(tallyFile)
            samplesInFile = list(data.keys())
            if 'input' in samplesInFile: # this is for compatibility with OpenMC runs on Bebop that were done before pytables was installed on the flux-up conda env on Bebop.
                samplesInFile.remove('input')
            [samplesList.append(sample) for sample in samplesInFile]
            [fileNamesList.append(tallyFile) for idx in samplesInFile]
        return fileNamesList,samplesList


class RelativeErrorTallyGenerator(TallyGenerator):

    def __call__(self):
        for file, sample in self.pairs:
            data = h5py.File(file)
            tallyX = moveaxis([data[sample][self.tallyNames[0]][:]])
            tallyY = moveaxis([data[sample][self.tallyNames[1]][:]])
            tallyXRelErr = moveaxis([data[sample][self.tallyNames[2]][:]])
            tallyYRelErr = moveaxis([data[sample][self.tallyNames[3]][:]])
            yield tallyX, tallyY, tallyXRelErr, tallyYRelErr

class splitTensorTallyGenerator(TallyGenerator):
    '''
    NB As of Jan 07 2024, this class is not compatible with most of the Pipeline
    classes because the data flow goes through TallyGenerator classes BEFORE
    the Pipeline. But because the normalization occurs in the Pipeline class,
    splitting the Tensors from the start causes incorrect normalization further
    down the flow.
    '''

    def __call__(self):
        for file, sample in self.pairs:
            data = h5py.File(file)
            tallyX = moveaxis([data[sample][self.tallyNames[0]][:]])
            volumetricNormFactor = np.shape(tallyX)[1]*np.shape(tallyX)[2]
            tallyX = tallyX*volumetricNormFactor
            tallyX = self.split_array(tallyX)
            tallyY = moveaxis([data[sample][self.tallyNames[1]][:]])
            tallyY = tallyY*volumetricNormFactor
            tallyY = self.split_array(tallyY,split_size=(40,40))
            tallyXRelErr = self.split_array(moveaxis([data[sample][self.tallyNames[2]][:]]))
            tallyYRelErr = self.split_array(moveaxis([data[sample][self.tallyNames[3]][:]]),split_size=(40,40))
            for i in range(0,len(tallyX)):
                yield tallyX[i], tallyY[i], tallyXRelErr[i], tallyYRelErr[i]
                #yield np.squeeze(tallyX[i]), np.squeeze(tallyY[i]), np.squeeze(tallyXRelErr[i]), np.squeeze(tallyYRelErr[i])

    def split_array(self, arr: np.array, split_size:tuple=(20,20)):
        shape = np.shape(arr)
        splits_needed = [shape[i] // split_size[i-1] for i in range(1, 3)]  # Assuming 20x20 for dimensions 1 and 2
            # Slice the array to create smaller arrays of shape (1,20,20,16)
        split_arrays = []
        for i in range(splits_needed[0]):
            for j in range(splits_needed[1]):
                sliced_array = arr[:, i * split_size[0] : (i + 1) * split_size[0], j * split_size[1] : (j + 1) * split_size[1], :]
                split_arrays.append(sliced_array)
        return split_arrays