
import openmc
import numpy as np
from sklearn.cluster import FeatureAgglomeration
import sys,os
from beavrs import materials
from iapws import IAPWS97

# Define materials
fuel = openmc.Material(name='Fuel')#Testing Testing 123
fuel.set_density('g/cm3', 10.29769)
fuel.add_nuclide('U234', 4.4843e-6)
fuel.add_nuclide('U235', 5.5815e-4)
fuel.add_nuclide('U238', 2.2408e-2)
fuel.add_nuclide('O16', 4.5829e-2)

clad = openmc.Material(name='Cladding')
clad.set_density('g/cm3', 6.55)
clad.add_nuclide('Zr90', 2.1827e-2)
clad.add_nuclide('Zr91', 4.7600e-3)
clad.add_nuclide('Zr92', 7.2758e-3)
clad.add_nuclide('Zr94', 7.3734e-3)
clad.add_nuclide('Zr96', 1.1879e-3)

hot_water = openmc.Material(name='Hot borated water')
hot_water.set_density('g/cm3', 0.740582)
hot_water.add_nuclide('H1', 4.9457e-2)
hot_water.add_nuclide('O16', 2.4672e-2)
hot_water.add_nuclide('B10', 8.0042e-6)
hot_water.add_nuclide('B11', 3.2218e-5)
hot_water.add_s_alpha_beta('c_H_in_H2O')

def assembly_model(sampleData,customData=None):
    """Returns a single PWR fuel assembly."""
    
    fuelRadius = sampleData['fuel_radius']
    model = openmc.model.Model()
    model.beavrsMaterials = materials.openmc_materials(1.0)
    assemblyPitch=21.42
    model.assemblyPitch = assemblyPitch
    # Create fuel assembly Lattice
    assembly = openmc.RectLattice(name='Fuel Assembly')
    assembly.pitch = (assemblyPitch/17, assemblyPitch/17)
    assembly.lower_left = (-assemblyPitch/2, -assemblyPitch/2)

    # fuelRadius in the dataframe can be a scalar, or a 17x17 array.
    fuelRadius = np.multiply(np.ones([17,17]),fuelRadius)

    # Create 17x17 array of universes for the sample assembly. We pass the configuration we want to generic_assembly
    # using the sampleData['assembly_materials'] object as an argument, and we pass the BEAVRS material library data 
    # using model.beavrsMaterials as an argument.
    assembly.universes = generic_assembly(fuelRadius,sampleData['assembly_materials'],model.beavrsMaterials)

    # Create outer boundary of the geometry to surround the lattice
    outer_boundary = openmc.model.rectangular_prism(
        assemblyPitch, assemblyPitch, boundary_type='reflective')

    # Create a cell filled with the lattice
    main_cell = openmc.Cell(fill=assembly, region=outer_boundary)

    # Finally, create geometry by providing a list of cells that fill the root
    # universe
    model.geometry = openmc.Geometry([main_cell])

    model.settings.batches = 150
    model.settings.inactive = 50
    model.settings.particles = 100000
    model.settings.source = openmc.Source(space=openmc.stats.Box(
        (-assemblyPitch/2, -assemblyPitch/2, -1),
        (assemblyPitch/2, assemblyPitch/2, 1),
        only_fissionable=True
    ))
    # the geometry plotting functions need to know what materials are in the assembly model. The model is returned by this method
    # so we store the openmc material definitions in the model before returning.
    model.hot_water = hot_water
    model.fuel = fuel
    model.clad = clad
    # NOTE: We never actually created a Materials object. When you export/run
    # using the Model object, if no materials were assigned it will look through
    # the Geometry object and automatically export any materials that are
    # necessary to build the model.
    return model

def generic_assembly(fuelRadius,materialIDs,beavrsMaterials):
    
    universeContainer = np.full((17,17),openmc.Universe)
    universes = {}
    for matName in beavrsMaterials:
        universes[matName] = openmc.Universe(name= matName + ' Pin')
    for i in range(0,17):
        for j in range(0,17):
            pin_or = openmc.ZCylinder(r=fuelRadius[i][j], name='Fuel OR')
            clad_or = openmc.ZCylinder(r=fuelRadius[i][j]+0.06502, name='Clad OR')
            clad_cell = openmc.Cell(fill=clad,region=+pin_or & -clad_or)
            hot_water_cell = openmc.Cell(fill=hot_water, region=+clad_or)
            pinMaterial = beavrsMaterials[materialIDs[i][j]]
            mat_cell = openmc.Cell(fill=pinMaterial,region=-pin_or)
            universeContainer[i,j] = universes[materialIDs[i][j]]
            universeContainer[i,j].add_cells([mat_cell,clad_cell,hot_water_cell])
    return universeContainer

def fuel_pin(fuel_or,clad_or):

    """Returns a fuel pin universe."""
    fuel_cell = openmc.Cell(fill=fuel, region=-fuel_or)
    clad_cell = openmc.Cell(fill=clad, region=+fuel_or & -clad_or)
    hot_water_cell = openmc.Cell(fill=hot_water, region=+clad_or)

    univ = openmc.Universe(name='Fuel Pin')
    univ.add_cells([fuel_cell, clad_cell, hot_water_cell])
    return univ


def guide_tube_pin(fuel_or,clad_or):
    """Returns a control rod guide tube universe."""

    gt_inner_cell = openmc.Cell(fill=hot_water, region=-fuel_or)
    gt_clad_cell = openmc.Cell(fill=clad, region=+fuel_or & -clad_or)
    gt_outer_cell = openmc.Cell(fill=hot_water, region=+clad_or)

    univ = openmc.Universe(name='Guide Tube')
    univ.add_cells([gt_inner_cell, gt_clad_cell, gt_outer_cell])
    return univ

def assembly_model_multiparameter(sampleData,customData):
    """Return OpenMC assembly with depleted materials and a number of customizations. Use make_samples.random_sweep_multiparameter() to
    generate the assembly parameters.


    """
    depletionResults = customData['depleted_materials']
    model = openmc.model.Model()
    asmPositions = sampleData['assembly_positions']
    assemblyPitch=21.50364 # BEAVRS
    model.assemblyPitch = assemblyPitch
    #beavrsMaterialsDict = materials.openmc_materials(1.0)
    beavrsMaterialsDict = depletionResults
    # Create fuel assembly Lattice
    assembly = openmc.RectLattice(name='Fuel Assembly')
    assembly.pitch = (assemblyPitch/asmPositions, assemblyPitch/asmPositions)
    assembly.lower_left = (-assemblyPitch/2, -assemblyPitch/2)

    
    # Create asmPositions x asmPositions array of universes for the sample assembly.
    assembly.universes = multiparameter_assembly(sampleData,beavrsMaterialsDict,depletionResults,assemblyPitch)

    # Create outer boundary of the geometry to surround the lattice
    outer_boundary = openmc.model.rectangular_prism(
        assemblyPitch, assemblyPitch, boundary_type='reflective')

    # Create a cell filled with the lattice
    main_cell = openmc.Cell(fill=assembly, region=outer_boundary)

    # Finally, create geometry by providing a list of cells that fill the root
    # universe
    model.geometry = openmc.Geometry([main_cell])

    model.settings.batches = 150
    model.settings.inactive = 50
    model.settings.particles = 100000
    model.settings.temperature={'method':'interpolation'}
    model.settings.source = openmc.Source(space=openmc.stats.Box(
        (-assemblyPitch/2, -assemblyPitch/2, -1),
        (assemblyPitch/2, assemblyPitch/2, 1),
        only_fissionable=True
    ))
    # the geometry plotting functions need to know what materials are in the assembly model. The model is returned by this method
    # so we store the openmc material definitions in the model before returning.
    model.hot_water = hot_water
    model.fuel = fuel
    model.clad = clad
    model.beavrsMaterials = beavrsMaterialsDict
    # NOTE: We never actually created a Materials object. When you export/run
    # using the Model object, if no materials were assigned it will look through
    # the Geometry object and automatically export any materials that are
    # necessary to build the model.
    return model


def multiparameter_assembly(sampleData, beavrsMaterialsDict,depletionResults,assemblyPitch):
    asmPositions = sampleData['assembly_positions']
    pincellDiameter = np.divide(assemblyPitch,asmPositions)
    pinRadii = 0.5*np.multiply(sampleData['fuel_radius'], pincellDiameter)
    cladThickness = np.multiply(sampleData['clad_thickness'], pinRadii)
    universeContainer = np.full((asmPositions,asmPositions),openmc.Universe)
    #hot_water.temperature = sampleData['coolant_temperature']
    materialIDs = sampleData['assembly_materials']
    compressed_water_iapws = IAPWS97(T=sampleData['coolant_temperature'],P=15.5)
    beavrsMaterialsDict['Borated Water'].temperature = sampleData['coolant_temperature']
    beavrsMaterialsDict['Pure Water'].temperature = beavrsMaterialsDict['Borated Water'].temperature
    beavrsMaterialsDict['Borated Water'].set_density('g/cm3', compressed_water_iapws.rho/1000)
    beavrsMaterialsDict['Pure Water'].set_density('g/cm3',beavrsMaterialsDict['Borated Water'].density)
    #hot_water.set_density('g/cm3', compressed_water_iapws.rho/1000)
    asmCoolantMaterial = beavrsMaterialsDict[sampleData['assembly_coolant_material']]
    for i in range(0,asmPositions):
        for j in range(0,asmPositions):
            pin_or = openmc.ZCylinder(r=pinRadii[i][j], name='Fuel OR')
            clad_or = openmc.ZCylinder(r=pinRadii[i][j]+cladThickness[i][j], name='Clad OR')
            clad_cell = openmc.Cell(fill=clad,region=+pin_or & -clad_or)
            hot_water_cell = openmc.Cell(fill=asmCoolantMaterial, region=+clad_or)
            
            pinMaterial = get_pin_material(beavrsMaterialsDict,depletionResults,materialIDs[i][j],sampleData['burnup_step'][i][j],sampleData)
            pinMaterial.temperature = sampleData['pin_temperature'][i][j]
            # TO-DO: see make_samples.random_sweep_multiparameter(): water pins always defined and named borated water but we override here
            # to always be the same as the coolant water
            if('Water' in materialIDs[i][j] or materialIDs[i][j] == 'B4C'):
                pinMaterial = asmCoolantMaterial
                pinMaterial.temperature = sampleData['coolant_temperature']
            if('Water' in materialIDs[i][j]):
                pinMaterial = asmCoolantMaterial
                #pinMaterial.set_density('g/cm3', asmCoolantMaterial.density)
            mat_cell = openmc.Cell(fill=pinMaterial,region=-pin_or)
            universe = openmc.Universe(name = materialIDs[i][j] + ' numID:' + str(pinMaterial.id))
            universeContainer[i,j] = universe
            universeContainer[i,j].add_cells([mat_cell,clad_cell,hot_water_cell])
    return universeContainer

def get_pin_material(beavrsMaterialsDict,depletionResults,materialName,burnupStep,sampleData):
    if(materialName  == 'Coolant Water'):
        return beavrsMaterialsDict[sampleData['assembly_coolant_material']]
    if(materialName == 'B4C'):
        return beavrsMaterialsDict[materialName]
    return depletionResults[materialName][burnupStep]
    