#! coding:utf-8

""" Data pipeline with Tensorflow data API.

"""

# Generic import
import os
import h5py
import numpy as np
import json

# Tensorflow
#os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import tensorflow as tf


def read_hdf5_results(path):
    """Read tallies results from HDF5 file.

    Parameters
    ----------
    path: str
        Path to the HDF5 file.

    Returns
    -------
    low_fidelity_tallies: np.array
        Low-fidelity tallies.
    high_fidelity_tallies: np.array
        High-fidelity tallies.

    """
    f = h5py.File(path, "r")
    low_fidelity_tallies = np.array(f['lo_fi'][:])
    low_fidelity_tallies = low_fidelity_tallies.reshape(
        list(low_fidelity_tallies.shape)+[1]
    )
    high_fidelity_tallies = np.array(f['hi_fi'][:])
    high_fidelity_tallies = high_fidelity_tallies.reshape(
        list(high_fidelity_tallies.shape)+[1]
    )
    return low_fidelity_tallies, high_fidelity_tallies


def make_tfrecord_from_hdf5(folder_paths):
    """ Create a TFRecord file from multiple hdf5 files.

    Parameters
    ----------
    folder_paths: list
        List of paths to folder results where an hdf5 file is stored.

    """
    # Reading tallies
    for i, folder in enumerate(folder_paths):
        result_path = os.path.join(folder, "result.hdf5")
        if i == 0:
            lofi_tallies, hifi_tallies = read_hdf5_results(result_path)
            shapes = (lofi_tallies[0].shape, hifi_tallies[0].shape)
        else:
            lofi_res, hifi_res = read_hdf5_results(result_path)
            shapes_res = (lofi_tallies[0].shape, hifi_tallies[0].shape)
            if shapes_res != shapes:
                print(f"Shape mismatch! Results from {folder} not added!")
                continue
            lofi_tallies = np.concatenate(
                (lofi_tallies, lofi_res),
                axis=0
            )
            hifi_tallies = np.concatenate(
                (hifi_tallies, hifi_res),
                axis=0
            )

    # Create examples
    examples = []
    for i in range(lofi_tallies.shape[0]):
        example = tf.train.Example(features=tf.train.Features(feature={
            "lofi": tf.train.Feature(
                float_list=tf.train.FloatList(
                    value=lofi_tallies[i].flatten()
                )
            ),
            "lofi_dim": tf.train.Feature(
                int64_list=tf.train.Int64List(
                    value=lofi_tallies.shape[1:]
                )
            ),
            "hifi": tf.train.Feature(
                float_list=tf.train.FloatList(
                    value=hifi_tallies[i].flatten()
                )
            ),
            "hifi_dim": tf.train.Feature(
                int64_list=tf.train.Int64List(
                    value=hifi_tallies.shape[1:]
                )
            )
        }))
        examples.append(example)

    # Write TFrecord file
    with tf.io.TFRecordWriter('data.tfrecord') as writer:
        for example in examples:
            writer.write(example.SerializeToString())

    # Write json metadata
    metadata = {
        "input_shape": shapes[0],
        "lofi_record_size": lofi_tallies[0].size,
        "hifi_record_size": hifi_tallies[0].size

    }
    with open("data.json", "w") as fp:
        json.dump(metadata, fp, indent=4)

    return


def parse_example(record, lofi_record_size, hifi_record_size):
    """ Parse a single TFRecord example.

    Parameters
    ----------
    record: tfrecord
        TFRecord example.
    lofi_record_size: int
        Size of lofi entry (from metadata file).
    hifi_record_size: int
        Size of hifi entry (from metadata file).

    Returns
    -------
    lofi: tf.Tensor
        Low-fidelity tensor for one TFRecord example.
    hifi: tf.Tensor
        High-fidelity tensor for one TFRecord example.

    """
    features = {
        "lofi": tf.io.FixedLenFeature([lofi_record_size], tf.float32),
        "lofi_dim": tf.io.FixedLenFeature([4], tf.int64),
        "hifi": tf.io.FixedLenFeature([hifi_record_size], tf.float32),
        "hifi_dim": tf.io.FixedLenFeature([4], tf.int64),
    }

    example = tf.io.parse_single_example(record, features)
    lofi = example["lofi"]
    lofi_dim = example["lofi_dim"]
    lofi = tf.reshape(lofi, lofi_dim)
    hifi = example["hifi"]
    hifi_dim = example["hifi_dim"]
    hifi = tf.reshape(hifi, hifi_dim)

    return lofi, hifi


def normalize_dataset(low, high):
    """Normalization function for the tensorflow
    data pipeline.

    Parameters
    ----------
    low: tf.Tensor
        Low-fidelity tensor for one sample.
    high: tf.Tensor
        Associated high-fidelity tensor.

    Returns
    -------
    low_norm: tf.Tensor
        Normalized low-fidelity tensor.
    high_norm: tf.Tensor
        Normalized high-fidelity tensor.

    """
    def normalize(sample):
        """Normalize one sample.

        """
        max = tf.reduce_max(sample)
        sample = tf.divide(sample, max)
        return sample

    low_norm = normalize(low)
    high_norm = normalize(high)

    return low_norm, high_norm


def create_pipeline_from_tfrecord(train_percent, batch_size):
    """ Create a data pipeline from existing TFRecord file.

    The TFRecord file must be named data.tfrecord and a json
    file named data.json must be present.

    Parameters
    ----------
    train_percent: int
        Fraction of example used for training. The remaining is
        used to calculate the validation score. Must be in [0,1].
    batch_size: int
        Size of one training data batch. Also used for validation
        batch size.

    Returns
    -------
    input_shape:
        Shape of one low-fidelity example.
    train_dataset: tf.data.Dataset
        Training dataset.
    val_dataset: tf.data.Dataset
        Validation dataset.
    """
    # Loading metadata from json
    with open("data.json", "r") as fp:
        metadata = json.load(fp)

    input_shape = metadata["input_shape"]
    lofi_record_size = metadata["lofi_record_size"]
    hifi_record_size = metadata["hifi_record_size"]

    # Creating data pipeline
    dataset = tf.data.TFRecordDataset(filenames = [
        'data.tfrecord'
    ])

    dataset_size = dataset.reduce(
        np.int64(0),
        lambda x, _:  x + 1
    ).numpy()

    dataset = dataset.map(
        lambda x: parse_example(
            x,
            lofi_record_size,
            hifi_record_size
        ),
        num_parallel_calls=tf.data.AUTOTUNE
    )

    dataset = dataset.map(
        normalize_dataset,
        num_parallel_calls=tf.data.AUTOTUNE
    )

    dataset = dataset.shuffle(dataset_size)

    train_size = round(dataset_size * train_percent)

    train_dataset = dataset.take(train_size)
    val_dataset = dataset.skip(train_size)

    train_dataset = train_dataset.batch(
        batch_size,
        num_parallel_calls=tf.data.AUTOTUNE
    )
    val_dataset = val_dataset.batch(
        batch_size,
        num_parallel_calls=tf.data.AUTOTUNE
    )
    return input_shape, train_dataset, val_dataset



def init_data_pipeline(folder_paths, train_percent=0.8, batch_size=10):
    """ Initialize a data pipeline from multiple results folder
    containing hdf5 files.

    This function will create a TFRecord file from the list of
    hfd5 files given.

    Parameters
    ----------
    folder_paths: list
        List of paths to folder results where an hdf5 file is stored.
    train_percent: int
        Fraction of example used for training. The remaining is
        used to calculate the validation score. Must be in [0,1].
    batch_size: int
        Size of one training data batch. Also used for validation
        batch size.

    Returns
    -------
    input_shape:
        Shape of one low-fidelity example.
    train_dataset: tf.data.Dataset
        Training dataset.
    val_dataset: tf.data.Dataset
        Validation dataset.

    """
    # Verification
    if train_percent < 0 and train_percent > 1:
        raise ValueError(
            f"train_percent must be in [0,1]: was {train_percent}"
        )

    # Transform hdf5 into TFRecord file
    make_tfrecord_from_hdf5(folder_paths)

    # Create pipeline
    input_shape, train_dataset, val_dataset = create_pipeline_from_tfrecord(
        train_percent,
        batch_size   
    )

    return input_shape, train_dataset, val_dataset
