import pandas as pd
import os,sys
import numpy as np
import beavrs.builder

def simple(Nsamples):
    fuelRadii = np.linspace(0.2,0.4,Nsamples)
    data = {'fuel_radius':fuelRadii,'assembly_positions':[pd.DataFrame(columns=['position','fuel_radius','custom_material']) for _ in range(Nsamples)]}
    df = pd.DataFrame(data)
    return df


def simpleRandom(Nsamples,seed=12345):
    choices = ['Borated Water','B4C','Fuel 1.6%','Fuel 2.4%','Fuel 3.1%','Fuel 3.2%','Fuel 3.4%']
    fuelRadius = 0.39218
    df = pd.DataFrame(columns=['fuel_radius','assembly_materials'])
    np.random.seed(seed) # make results reproducible
    for i in range(0,Nsamples):
        datai = pd.DataFrame({'fuel_radius':[fuelRadius],'assembly_materials':[np.random.choice(choices, size=(17, 17))]})
        df = pd.concat([df,datai])
    return df      
        
def simpleRandomAssignRadii(Nsamples,seed=12345,radii=[0.39218,0.39218,0.39218]):
    D = {'Borated Water':radii[0],'B4C':radii[1],'Fuel 1.6%':radii[2],'Fuel 2.4%':radii[2],'Fuel 3.1%':radii[2],'Fuel 3.2%':radii[2],'Fuel 3.4%':radii[2]}
    choices = list(D.keys())            
    df = pd.DataFrame(columns=['fuel_radius','assembly_materials'])
    np.random.seed(seed) # make results reproducible
    for i in range(0,Nsamples):
        assemblyMats=np.random.choice(choices, size=(17, 17))
        ss=np.shape(assemblyMats)
        fuelRadii=np.zeros(ss)
        for i in range(0,ss[0]):
            for j in range(0,ss[1]):
                fuelRadii[i][j]=D[assemblyMats[i][j]]
        datai = pd.DataFrame({'fuel_radius':[fuelRadii],'assembly_materials':[assemblyMats]})
        df = pd.concat([df,datai])
    return df   


def BEAVRS_assemblies():
    """Create DataFrame with BEAVRS assembly (lattice) names and objects. See BEAVRS example extract-assm.ipynb. 

    Parameters
    ----------
    None

    Returns
    -------
    df: Pandas DataFrame
        df contains the lattice names and objects corresponding to names used in BEAVRS unique assembly model types 
    """
    df = pd.DataFrame(columns=['lattice_name','lattice_object'])
    b = beavrs.builder.BEAVRS()
    all_latts = b.openmc_geometry.get_all_lattices()
    for id,latt in all_latts.items():
        if latt.name == 'Core Lattice': # the core lattice is not desired here - only assemblies.
            continue
        datai = pd.DataFrame({'lattice_name':[latt.name],'lattice_object':[latt]})
        df = pd.concat([df,datai])
    return df


from skopt.space import Space
from skopt.sampler import Lhs
def random_sweep_multiparameter(
    Nsamples, 
    seed=12345, 
    diamMinMax=(0.2,0.8), 
    assemblyPositionsMinMax=(15,19),
    burnupStepChoices = (0,12), # all fuel enrichments are burned from 0 to 12 steps. BU levels vary by enrichment.
    cladThicknessChoices = (0.1,0.2), # cladding should be between 0.1 and 0.2 of the radius of the overall pin
    fuelTchoices = (350.,2000.),
    coolantTchoices = (350.,566.), # NB this is AVERAGE coolant T not exit coolant T.
    fuelEnrichmentChoices = ['Fuel 1.6%','Fuel 2.4%','Fuel 3.1%','Fuel 3.2%','Fuel 3.4%']
    ):
    """Create DataFrame whose Nsamples rows each describe a fuel assembly.
    Assembly parameters (temperatures, pin radii, lattice elements, matrials, burnup) are randomly sampled
    using Latin Hypercube sampling

    Scikit-Optimize libary required, install using conda install -c conda-forge scikit-optimize

    Parameters
    ----------
    Nsamples: int, scalar, number of samples (assemblies) to generate
    seed: scalar, integer, random number seed for reproducibility.
    diamMinMax: tuple of (lower, upper) pin diameters expressed as a fraction of a unit pincell's width (or pin pitch).
    assemblyPositionsMinMax: tuple of (lower, upper) integer assembly positions: e.g. (15,19) gives 15x15 to 19x19 assemblies.
    burnupStepChoices: tuple of (lower, upper) integer burnup step choices
    cladThicknessChoices: tuple of (lower,upper) clad thicknesses expressed as a fraction of a pin's radius
    fuelTchoices: tuple of (lower, upper) real numbers, temperatures in K of fuel pins (not B4C or borated water pins)
    coolantTchoices: tuple of (lower, upper) real numbers, temperatures in K of coolant. Applies to whole assembly, and all B4C and borated water positions.
    fuelEnrichmentChoices: list of material names corresponding to the enrichments of fuel available and desired.
    Returns
    -------
    df: Pandas DataFrame
        df contains Nsamples assemblies generated randomly using latin hypercube sampling.
    """
    np.random.seed(seed) # make results reproducible
    dfColumns = ['fuel_radius','assembly_materials', 'assembly_positions', 'coolant_temperature', 
        'burnup_step', 'clad_thickness', 'pin_temperature']
    df = pd.DataFrame(columns=dfColumns)
    pinMterialChoices = ['Coolant Water','B4C']
    pinMterialChoices.extend(fuelEnrichmentChoices) # add provided fuel enrichments on the end
    
    
    
    # whole-assembly parameters: coolant T and num of assembly positions
    #lhs = Lhs(criterion="maximin", iterations=10000)
    lhs = Lhs(lhs_type="classic", criterion=None)
    #asmSpace = Space([coolantTchoices,assemblyPositionsMinMax])
    asmSpace = Space([coolantTchoices,assemblyPositionsMinMax,['Pure Water','Borated Water']])
    asmParameters = lhs.generate(asmSpace.dimensions, Nsamples)
    pinSpace = Space([diamMinMax,pinMterialChoices,burnupStepChoices,cladThicknessChoices,fuelTchoices])
   
    for i in range(0,Nsamples):
        asmPositions = asmParameters[i][1]
        nPins = asmPositions**2
        pinParameters = np.array(lhs.generate(pinSpace.dimensions, n_samples=nPins))
        pinRadii = np.reshape(np.float64(pinParameters[:,0]),(asmPositions,asmPositions))
        pinMaterials = np.reshape(pinParameters[:,1],(asmPositions,asmPositions))
        pinBUStep = np.reshape(np.int32(pinParameters[:,2]),(asmPositions,asmPositions))
        pinCladThickness = np.reshape(np.float64(pinParameters[:,3]),(asmPositions,asmPositions))
        pinTemp = np.reshape(np.float64(pinParameters[:,4]),(asmPositions,asmPositions))
        datai = pd.DataFrame({'fuel_radius':[pinRadii], 'assembly_materials':[pinMaterials],
            'coolant_temperature':[asmParameters[i][0]], 'assembly_positions': [asmParameters[i][1]],
            'burnup_step':[pinBUStep],'clad_thickness':[pinCladThickness],
            'pin_temperature':[pinTemp],'assembly_coolant_material':[asmParameters[i][2]]})
        df = pd.concat([df,datai])
    return df

import src.utilities as ut
def random_sweep_multiparameter_SFR(
    Nsamples,
    seed=12345,
    numRowsMinMax = (2,11),
    pinDiameterMinMax = (0.5, 1.0),
    pinPitchMinMax = (0.51,1.32),
    ductThicknessMinMax = (0,0.24),
    fuelTypeChoices = (0,1), # NB the fuel type is defined by first dimension of the ID array generated by AssemblyModel.populateDepletedMaterials()
    burnupStepChoices = (0,9), # NB the burnup step is defined by second dimension of the ID array generated by AssemblyModel.populateDepletedMaterials()
    cladThicknessMinMax= (0.0633,0.0942), # cladding should be between 0.0633 and 0.0942 of the radius of the overall pin
    fuelTminMax = (350.,2000.),
    coolantTminMax = (350.,566.), # NB this is AVERAGE coolant T not exit coolant T.
    assemblyMaskOverrideChoices = [None,None] # insert e.g. control rod or empty rod in place of fuel.     
    ) -> pd.DataFrame:
    '''Create DataFrame whose Nsamples rows each describe a fuel assembly.
    Assembly parameters (temperatures, pin radii, lattice elements, matrials, burnup) are randomly sampled
    using Latin Hypercube sampling

    Scikit-Optimize libary required, install using conda install -c conda-forge scikit-optimize

    Parameters
    ----------
    Nsamples: int, scalar, number of samples (assemblies) to generate
    seed: scalar, integer, random number seed for reproducibility.
    numRowsMinMax: tuple of (lower, upper) pin rows (see hex asm definition in Waltar & Todd)
    pinDiameterMinMax: tuple of (lower, upper) OUTER pin diameters in units of fractions of pin pitch 
    pinPitchMinMax: tuple of (lower, upper) real number expressing pin pitch limits in cm
    fuelTypeChoices: tuple containing choices of fuel enrichment integer ID numbers
    burnupStepMinMax: tuple of (lower, upper) integer burnup step choices
    cladThicknessMinMax: tuple of (lower,upper) clad thicknesses expressed as a fraction of a pin's radius
    fuelTminMax: tuple of (lower, upper) real numbers, temperatures in K of fuel pins (not B4C or borated water pins)
    coolantTminMax: tuple of (lower, upper) real numbers, temperatures in K of coolant. Applies to whole assembly, and all B4C and borated water positions.
    assemblyMaskOverrideChoices: tuple of choices for replacing fuel. E.g. None = no override, 'B4C' = control rod, 'Empty' = sodium. NB the specific definition of each choice is determined by the assembly model code.
    Returns
    -------
    df: Pandas DataFrame
        df contains Nsamples assemblies generated randomly using latin hypercube sampling.'''
    np.random.seed(seed) # make results reproducible
    dfColumns = ['num_rows','pin_radius', 'pin_pitch',  'duct_thickness', 'fuel_type',
        'burnup_step', 'clad_thickness', 'pin_temperature', 'coolant_temperature','mask_overrides']
    df = pd.DataFrame(columns=dfColumns)
    lhs = Lhs(lhs_type="classic", criterion=None)
    asmSpace = Space([numRowsMinMax,ductThicknessMinMax,coolantTminMax,pinPitchMinMax])
    asmParameters = lhs.generate(asmSpace.dimensions, Nsamples)
    pinSpace = Space([pinDiameterMinMax,fuelTypeChoices,burnupStepChoices,cladThicknessMinMax,fuelTminMax])
    for i in range(0,Nsamples):
        nRows = asmParameters[i][0]
        pinPitch = asmParameters[i][3]
        nPinsTotal = ut.total_pins_hex(nRows)
        pinParameters = lhs.generate(pinSpace.dimensions,n_samples=nPinsTotal)
        pinRadii = 0.5*pinPitch*np.array([par[0] for par in pinParameters])
        #pinRadii[2*pinRadii > pinPitch] = pinPitch/2
        pinMaterials = np.array([par[1] for par in pinParameters])
        pinBUStep = np.array([par[2] for par in pinParameters])
        pinCladThickness = np.array([par[3] for par in pinParameters])
        pinTemp =np.array([par[4] for par in pinParameters])
        maskOverrides = np.random.choice(assemblyMaskOverrideChoices,size=nPinsTotal)
        datai = pd.DataFrame({'num_rows':nRows,'pin_radius':[pinRadii],'pin_pitch':pinPitch,'clad_thickness':[pinCladThickness],
            'duct_thickness':asmParameters[i][1],'pin_temperature':[pinTemp],'coolant_temperature':asmParameters[i][2],
            'fuel_type':[pinMaterials],'burnup_step':[pinBUStep],'mask_overrides':[maskOverrides]})
        df = pd.concat([df,datai])
    return df
    

    
    