import numpy as np

# samples is modified-in-place (pass by reference)
def normalize_per_sample(samples,returnScalingFactor=False):
    Nsamples = samples.shape[0]
    theMax = np.zeros((Nsamples,))
    for i in np.arange(0,Nsamples):
        allData = np.ravel(samples[i,:,:,:])
        theMax[i] = np.max(allData)
        samples[i,:,:,:] = np.divide(samples[i,:,:,:], theMax[i])
    if(returnScalingFactor):
        return samples,theMax 
    else: 
        return samples

# samples is modified-in-place (pass by reference)
def normalize_by_group(samples,returnScalingFactor=False):
    numberSamples = samples.shape[0]
    numberGroups = samples.shape[1]
    maxFlux = np.zeros((numberSamples,numberGroups))
    for i in np.arange(0,numberSamples):
        for j in np.arange(0,numberGroups):
            flatData = np.ravel(samples[i,j,:,:])
            maxFlux[i,j] = np.max(flatData)
            samples[i,j,:,:] = np.divide(samples[i,j,:,:], maxFlux[i,j])
    if(returnScalingFactor):
        return samples,maxFlux 
    else: 
        return samples

def normalize_by_group_only(samples,returnScalingFactor=False,scalingFactors=None):
    """ Scale samples according to max values in individual groups.

    Parameters
    ----------
    samples: numpy ndarray
        4-D ndarray (Nsamples,Ngroups,Nx,Ny)
    returnScalingFactor: boolean
        Set to true to return the scaling factor for each group.
    scalingFactors: None or np.ndarray
        Pass a 1-D np.ndarray (Ngroups) to force the function to use those to scale the samples.
        Setting to None (default) causes the function to find the scaling factors.

    Returns
    -------
    samples: np.ndarray
        4-D ndarray (Nsamples,Ngroups,Nx,Ny) of scaled values.
    scalingFactors: np.ndarray
        1-D np.ndarray (Ngroups) of scaling factors found.    
    """
    numberGroups = samples.shape[1]
    #if np.any(scalingFactors, None):
    if scalingFactors is None:
        scalingFactors = np.zeros((numberGroups,1))
        for j in np.arange(0,numberGroups):
            flatData = np.ravel(samples[:,j,:,:])
            scalingFactors[j] = np.max(flatData)
            samples[:,j,:,:] = np.divide(samples[:,j,:,:], scalingFactors[j])
    else:
        for j in np.arange(0,numberGroups):
            flatData = np.ravel(samples[:,j,:,:])
            samples[:,j,:,:] = np.divide(samples[:,j,:,:], scalingFactors[j])   
    if(returnScalingFactor):
        return samples,scalingFactors 
    else: 
        return samples 

def normalize_one_group(samples,returnScalingFactor=False,scalingFactor=None):
    """ Scale samples of one-group data according to max value in all data.  

    Parameters
    ----------
    samples: numpy ndarray
        3-D ndarray (Nsamples,Nx,Ny)
    returnScalingFactor: boolean
        Set to true to return the scaling factor.
    scalingFactor: None or np.ndarray
        Pass a scalar to force the function to use it to scale the samples.
        Setting to None (default) causes the function to find the scaling factor.

    Returns
    -------
    samples: np.ndarray
        3-D ndarray (Nsamples,Nx,Ny) of scaled values.
    scalingFactor: np.ndarray
        scaling factor found.
    """
    flatData = np.ravel(samples)
    if scalingFactor is None:
        scalingFactor = np.max(flatData)    
    samples = np.divide(samples, scalingFactor)   
    if(returnScalingFactor):
        return samples,scalingFactor
    else: 
        return samples


# works like sklearn.preprocessing.StandardScaler: x_scaled = [x-mean(X)/std(X)]
def scale_by_group_stdev(samples,returnScalingFactor=False):
    numberSamples = samples.shape[0]
    numberGroups = samples.shape[1]
    mean = np.zeros((numberGroups,1))
    std = np.zeros((numberGroups,1))
    for j in np.arange(0,numberGroups):
        flatData = np.ravel(samples[:,j,:,:])
        mean[j] = np.mean(flatData)
        std[j] = np.std(flatData)
        samples[:,j,:,:] = np.divide(samples[:,j,:,:]-mean[j], std[j])
    if(returnScalingFactor):
        return samples,(mean,std) 
    else: 
        return samples     

from scipy.io import loadmat
def convenient_load_normalization_values(datapath):
    matDict = loadmat(datapath + '/out.mat', variable_names=['loFiScalingValue','hiFiScalingValue'])
    return matDict['loFiScalingValue'],matDict['hiFiScalingValue']


import tensorflow as tf
def spatial_max(gen,groupsLo=8,groupsHi=16):
    '''
    Searches over data produced by generator gen, finds the maximum value in each low and high resolution energy group.
    
    '''
    def max_sample(X,Y):
        return tf.reduce_max(X,axis=[0,1,2]), tf.reduce_max(Y,axis=[0,1,2])
    maxesX = np.empty(shape=[1,groupsLo],dtype=np.float32)
    maxesY = np.empty(shape=[1,groupsHi],dtype=np.float32)
    for result in gen:
        X = result[0]
        Y = result[1]
        maxX,maxY = max_sample(X,Y)
        maxesX = np.append(maxesX,np.reshape(np.array(maxX,dtype=np.float32),(1,groupsLo)),axis=0)
        maxesY = np.append(maxesY,np.reshape(np.array(maxY,dtype=np.float32),(1,groupsHi)),axis=0)
    maxesX = np.delete(maxesX,0,0)
    maxesY = np.delete(maxesY,0,0)
    spatialMaxX = tf.reshape(tf.reduce_max(maxesX,axis=0), [1,1,1,groupsLo])
    spatialMaxY = tf.reshape(tf.reduce_max(maxesY,axis=0), [1,1,1,groupsHi])
    return spatialMaxX,spatialMaxY


def spatial_mean(gen,groupsLo=8,groupsHi=16):
    '''
    Searches over data produced by generator gen, finds the mean value in each low and high resolution energy group.
    
    '''
    def mean_sample(X,Y):
        return tf.reduce_mean(X,axis=[0,1,2]), tf.reduce_mean(Y,axis=[0,1,2])
    meansX = np.empty(shape=[1,groupsLo],dtype=np.float32)
    meansY = np.empty(shape=[1,groupsHi],dtype=np.float32)
    for result in gen:
        X = result[0]
        Y = result[1]
        maxX,maxY = mean_sample(X,Y)
        meansX = np.append(meansX,np.reshape(np.array(maxX,dtype=np.float32),(1,groupsLo)),axis=0)
        meansY = np.append(meansY,np.reshape(np.array(maxY,dtype=np.float32),(1,groupsHi)),axis=0)
    meansX = np.delete(meansX,0,0)
    meansY = np.delete(meansY,0,0)
    spatialMeanX = tf.reshape(tf.reduce_mean(meansX,axis=0), [1,1,1,groupsLo])
    spatialMeanY = tf.reshape(tf.reduce_mean(meansY,axis=0), [1,1,1,groupsHi])
    return spatialMeanX,spatialMeanY