import h5py
import os
import numpy as np

def mergeTallies(pathList,includeUncertainty=False,flatten=False):
    """ Load tallies from multiple runs and concatenate them

    Parameters
    ----------
    pathList: list
        list of paths to OpenMC runs
    includeUncertainty: boolean
        Set to true to return the tally uncertainties
    flatten: boolean
        Set to true to sum N-group tallies, giving a one-group merged tally.
        NB This does not apply to uncertainties, which have to be flattened with error propagation.

    Returns
    -------
    loFiTallies: np.ndarray
        4-D ndarray (Nsamples,Ngroups,Nx,Ny) of low fidelity tally results (Nsamples,Nx,Ny) if flatten=True
    hiFiTallies: np.ndarray
        4-D ndarray (Nsamples,Ngroups,Nx,Ny) of high fidelity tally results (Nsamples,Nx,Ny) if flatten=True.
    loFiRelErr: np.ndarray
        4-D ndarray (Nsamples,Ngroups,Nx,Ny) of low fidelity relative error results (Nsamples,Nx,Ny) if flatten=True
    hiFiRelErr: np.ndarray
        4-D ndarray (Nsamples,Ngroups,Nx,Ny) of high fidelity relative error results (Nsamples,Nx,Ny) if flatten=True    
    loFiStd: np.ndarray
        4-D ndarray (Nsamples,Ngroups,Nx,Ny) of low fidelity standard deviation results (Nsamples,Nx,Ny) if flatten=True
    hiFiStd: np.ndarray
        4-D ndarray (Nsamples,Ngroups,Nx,Ny) of high fidelity standard deviation results (Nsamples,Nx,Ny) if flatten=True        
    """
    loFiTallies = np.array([])
    hiFiTallies = np.array([])
    loFiRelErr = np.array([])
    hiFiRelErr = np.array([])
    loFiStd = np.array([])
    hiFiStd = np.array([])
    loFiTallies = np.concatenate([get_tally(filePath,'lo_fi',flatten) for filePath in pathList],axis=0)
    hiFiTallies = np.concatenate([get_tally(filePath,'hi_fi',flatten) for filePath in pathList],axis=0)
    if includeUncertainty:
        loFiRelErr = np.concatenate([get_tally(filePath,'lo_fi_rel_err') for filePath in pathList],axis=0)
        hiFiRelErr = np.concatenate([get_tally(filePath,'hi_fi_rel_err') for filePath in pathList],axis=0)
        loFiStd = np.concatenate([get_tally_std(filePath,'lo_fi_std_dev',flatten) for filePath in pathList],axis=0)
        hiFiStd = np.concatenate([get_tally_std(filePath,'hi_fi_std_dev',flatten) for filePath in pathList],axis=0)
        if flatten:
            loFiRelErr = np.divide(loFiStd,loFiTallies)
            hiFiRelErr = np.divide(hiFiStd,hiFiTallies)
        return loFiTallies,hiFiTallies, loFiRelErr, hiFiRelErr, loFiStd, hiFiStd
    else:    
        return loFiTallies,hiFiTallies

def merge_named_tallies(pathList, talliesList):
    tallyDict = {}
    for tallyName in talliesList:
        tallyDict[tallyName] = np.concatenate([get_tally(filePath,tallyName,flatten=False) for filePath in pathList],axis=0)
    return tallyDict


def get_tally_list(filePath:str, tallyName:str) -> list:
    ''' Get tallies with variable shapes, e.g. those produced by SFROpenMCWrapper

    The function will automatically look for samples inside the file and pull out the tally according to the name given.

    Parameters
    ----------
    filePath: str
        path to relevant HDF5 tally results file
    tallyName: str
        Name of tally, e.g. "lo_fi", "hi_fi","lo_fi_rel_err" etc etc etc. 

    Returns
    ---------
    tally: list
        List of numpy.ndarrays
    '''
    file = os.path.join(filePath, 'result.hdf5')
    data = h5py.File(file)
    samplesList = list(data.keys())
    samplesList.remove('input') # HDF5 files produced by SFROpenMCWrapper also have an "input" group to store each sim's parameters.
    talliesList = []
    for groupName in samplesList:
        tally = data[groupName][tallyName][:]
        talliesList.append(tally)
    return talliesList

def get_tally(filePath,tallyName,flatten=False):
    file = os.path.join(filePath,'result.hdf5')
    data = h5py.File(file)
    tally = data[tallyName][:]
    if flatten:
        tally = np.sum(tally, 1) # sum along groups
    return tally

def get_tally_std(filePath,tallyName,flatten=False):
    tally = get_tally(filePath,tallyName)
    if flatten:
        tally = np.sqrt(np.sum(np.power(tally,2), 1)) # sum along groups
    return tally


# To-do: include uncertainties.
def flatten_tallies_uncertainties_groupwise(samples):
    """ Flatten tallies and their uncertainties in a groupwise fashion. E.g. tally(x,y,E) -> tally(x,y).

    Parameters
    ----------
    folder_paths: list
        List of paths to folder results where an hdf5 file is stored.

    """
    return None

def noisify_tallies_normal(tallyMeans,tallyStdevs,stdFac=1.0,seed=12345):
    """ Return a tally with added noise by randomly sampling from a normal distribution using the given tally mean and
    a factor by which to multiply the tally standard deviation.

    Parameters
    ----------
    tallyMeans:     [Nsamples x Ngroups x Nx x Ny] array of tally means (tally results)
    tallyStdevs:    [Nsamples x Ngroups x Nx x Ny] array of tally stdevs
    stdFac:         Scalar, factor by which to multiply the tallyStdevs. 
    seed:           Scalar, random number generator seed

    Returns
    ----------
    noisyTallyMeans: [Nsamples x Ngroups x Nx x Ny] array of tallies sampled from normal distribution with mean tallyMeans
                     and stdevs tallyStdevs*stdFac
    """
    rng = np.random.default_rng(seed)
    stdevs = np.multiply(tallyStdevs,stdFac)
    noisyTallyMeans = rng.normal(loc=tallyMeans,scale=stdevs)
    return noisyTallyMeans


def augment_tallies_8fold(a):
    """ Return original tallies with 7 "copies" appended: 1 copy is reflected about vertical axis,
    3 are rotated 90, 180, 270 degrees, and 3 that are the reflection rotated the same. See ANN lab book 04/24/23
    for logic.
    
    Parameters
    ----------
    a:     [Nsamples x Ngroups x Nx x Ny] array of tally results (can be tally means or stdevs etc.)
    
    Returns
    ----------
    augmentedTallies: [(8xNsamples) x Ngroups x Nx x Ny] array of augmented results.
    """
    b = np.flip(a, 3)
    # rotate original
    c = np.rot90(a, k=-1, axes=[2,3])
    d = np.rot90(c, k=-1, axes=[2,3])
    e = np.rot90(d, k=-1, axes=[2,3])
    # rotate reflection
    f = np.rot90(b, k=-1, axes=[2,3])
    g = np.rot90(f, k=-1, axes=[2,3])
    h = np.rot90(g, k=-1, axes=[2,3])
    augmentedTallies = np.concatenate([a,b,c,d,e,f,g,h],axis=0)
    return augmentedTallies

    """
    augmentedTallies = np.zeros((7*Nsamples,Ngroups,N,N)) # We'll append the originals at the end.
    for i in range(1,Nsamples):
        for j in range(1,Ngroups):
            b = np.reshape(np.fliplr(np.squeeze(tallies[i,j,:,:])),(1,1,N,N)) # reflect
            # rotate original
            c = np.rot90(tallies, k=-1, axes=[2,3])
            d = np.rot90(c, k=-1, axes=[2,3])
            e = np.rot90(c, k=-1, axes=[2,3])
            # rotate reflection
            f = np.rot90(b, k=-1, axes=[2,3])
            g = np.rot90(f, k=-1, axes=[2,3])
            h = np.rot90(g, k=-1, axes=[2,3])

    """
import tensorflow as tf
def augment_tallies_8fold_tf(a):
    """ Return original tallies with 7 "copies" appended: 1 copy is reflected about vertical axis,
    3 are rotated 90, 180, 270 degrees, and 3 that are the reflection rotated the same. See ANN lab book 04/24/23
    for logic.
    
    Parameters
    ----------
    a:     [Nsamples x   Nx x Ny x Ngroups] array of tally results (can be tally means or stdevs etc.)
    
    Returns
    ----------
    augmentedTallies: [(8xNsamples) x Ngroups x Nx x Ny] array of augmented results.
    """
    b = tf.image.flip_up_down(a)
    # rotate original
    c = tf.image.rot90(a, k=-1)
    d = tf.image.rot90(c, k=-1)
    e = tf.image.rot90(d, k=-1)
    # rotate reflection
    f = tf.image.rot90(b, k=-1)
    g = tf.image.rot90(f, k=-1)
    h = tf.image.rot90(g, k=-1)
    augmentedTallies = tf.stack([a,b,c,d,e,f,g,h],axis=0)
    #augmentedTallies = tf.data.Dataset.from_tensor_slices([a,b,c,d,e,f,g,h])
    return augmentedTallies    
# %%
def moveaxis(array4D):
    return np.moveaxis(array4D,[1,2,3],[3,1,2])

def unmoveaxis(array4D):
    return np.moveaxis(array4D,[3,1,2],[1,2,3])

#%% simplify prediction pipeline when using Gaussian-NLL-loss type models
def predict_identity(model,X):    
    X = moveaxis(X)
    theShape = np.shape(X) # (Samples, Nx,Ny,Ngroups)
    Ngroups = theShape[3]
    identIn = np.zeros(np.shape(X))
    y = model.predict([X,identIn])
    y_data = y[:,:,:,:Ngroups]
    return unmoveaxis(y_data)

#%% e.g. take two arrays (6400,8,64,64) each and create one (6400,16,64,64)
# right now works only for same shape arrays.
def merge_arrays_samples_to_groups(array1, array2):
    [Nsamples,Ngroups,Nx,Ny] = np.shape(array1)
    newArray = np.zeros((Nsamples,Ngroups*2,Nx,Ny))
    newArray[:,:Ngroups,:,:] = array1
    newArray[:,Ngroups:,:,:] = array2
    return newArray

#%% use sliding window to create N subsamples from inputArray of shape (Nsamples,Ngroups,Nx,Ny)
def subsample_sliding_window(inputArray,windowShape,stride):
    (Nsamples,Ngroups,Nx,Ny) = np.shape(inputArray)
    (m,n) = windowShape
    numSamplesRow = (Nx - m)//stride + 1
    numSamplesCol = (Ny - n)//stride + 1
    totalSubSamples = numSamplesRow*numSamplesRow
    subSamples = []
    for i in range(numSamplesRow):
        for j in range(numSamplesCol):
            subSampleij = inputArray[:,:,i*stride:i*stride+m,j*stride:j*stride+m]
            subSamples.append(subSampleij)
    
    subSamplesArray = np.array(subSamples)
    subSamplesArrayReshaped = np.reshape(subSamplesArray, (Nsamples*totalSubSamples,Ngroups,m,n))
    return subSamplesArrayReshaped

# this only works right now when subsamples are a complete subset of the original array... e.g. if
# you subsample a 64x64 array to 8x8 with stride 8.
#def unsubsample_sliding_window(inputArray,windowShape,stride):


#%% simplify prediction pipeline when using Gaussian-NLL loss type models and sliding window subsampling
def predict_subsamples_identity(model,X,windowShape=(8,8),stride=8):
    (Nsamples,Ngroups,Nx,Ny) = np.shape(X)
    y = np.zeros((Nsamples,Ngroups,Nx,Ny))
    (m,n) = windowShape
    for j in range(m):
        for k in range(n):
            y[:,:,j*stride:j*stride+m,k*stride:k*stride+n] = predict_identity(model,X[:,:,j*stride:j*stride+m,k*stride:k*stride+n])
    return y

#%% compute tally grid elements up to a max of 128x128 such that the delta-x and delta-y are always fixed
# e.g. grid cells should be 21.24cm/64 and 21.24cm/128 for low and high res tallies if we go by the original paper.
# returns a scalar N since the sampled region is an NxN square, corresponding to low-res tally. I.e. for high res multiply by 2.
# also returns the deltaXY value although it's a "constant". The reason is so that it doesn't get defined anywhere else.
def calculate_grid_elements_lores(boundingBoxSize, deltaXY=21.24/64):
    N = np.floor(boundingBoxSize/deltaXY) # round down, it's better to sample only the inside of an assembly and leave out a small part than attempt to sample nonphysical space
    if N>64:
        N = 64
    return N,deltaXY

def total_pins_hex(Nrows)->int:
    return int(6*(Nrows - 1)*(Nrows/2) + 1)

def print_to_file(object,filename:str) -> None:
    with open(filename, "w") as file: print(object, file=file)